"""
This file contains a wrapper class for converting annotations to multi-resolution image masks.
"""

from ..errors import imageerrors as dptimageerrors

import multiresolutionimageinterface as mir

import math

#----------------------------------------------------------------------------------------------------

class AnnotationConverter(object):
    """Wrapper class for annotation conversion."""

    def __init__(self, annotation_path, pixel_spacing):
        """
        Initialize the object and open the given annotation file.

        Args:
            annotation_path (str): Path of the annotation to load.
            pixel_spacing (float): Pixel spacing of the annotation (micrometer).

        Raises:
            AnnotationOpenError: The specified annotation cannot be opened for reading.
            InvalidPixelSpacingError: Invalid pixel spacing configuration.

        """

        # Initialize the base class.
        #
        super().__init__()

        # Initialize members.
        #
        self.__annotation = None  # Annotation object.
        self.__path = ''          # Path of the opened annotation.
        self.__spacing = None     # Pixel spacing of the loaded annotation.
        self.__groups = {}        # Annotation group to annotation count map.

        # Open annotation.
        #
        self.__openannotation(annotation_path=annotation_path)
        self.__setspacing(pixel_spacing=pixel_spacing)
        self.__collectgroups()

    def __openannotation(self, annotation_path):
        """
        Open annotation file.

        Args:
            annotation_path (str): Path of the annotation to load.

        Raises:
            AnnotationOpenError: The specified annotation cannot be opened for reading.
        """

        annotation = mir.AnnotationList()
        repository = mir.XmlRepository(annotation)
        repository.setSource(annotation_path)
        annotation_loaded = repository.load()

        # Check if annotation is successfully opened.
        #
        if not annotation_loaded:
            raise dptimageerrors.AnnotationOpenError(annotation_path)

        # Save the annotation object.
        #
        self.__annotation = annotation
        self.__path = annotation_path

    def __setspacing(self, pixel_spacing):
        """
        Set the pixel spacing and tolerance.

        Args:
            pixel_spacing (float): Pixel spacing of the annotation (micrometer).

        Raises:
            InvalidPixelSpacingError: Invalid pixel spacing configuration.
        """

        # The pixel spacing and the tolerance must be non-negative.
        #
        if pixel_spacing < 0.0:
            raise dptimageerrors.InvalidPixelSpacingError(self.__path, pixel_spacing)

        # Save the pixel spacing.
        #
        self.__spacing = pixel_spacing

    def __collectgroups(self):
        """Collect the annotation group information."""

        # Initialize the groups.
        #
        annotation_groups = self.__annotation.getGroups()
        for annotation_group_item in annotation_groups:
            self.__groups[annotation_group_item.getName()] = 0

        # Count the number of annotations in each group.
        #
        annotation_list = self.__annotation.getAnnotations()
        for annotation_item in annotation_list:
            self.__groups[annotation_item.getGroup().getName()] += 1

    def __rescaledannotations(self, pixel_spacing):
        """
        Rescale the annotations to the given pixel spacing.

        Args:
            pixel_spacing (float, None): Required pixel spacing (micrometer).
        """

        if pixel_spacing is not None and pixel_spacing <= 0.0:
            raise dptimageerrors.InvalidPixelSpacingError(self.__path, pixel_spacing)

        if pixel_spacing is not None and not math.isclose(pixel_spacing, self.__spacing):
            # Calculate the scaling factor.
            #
            factor = pixel_spacing / self.__spacing

            # Initialize the rescaled annotation list object and copy the groups.
            #
            rescaled_annotation = mir.AnnotationList()
            rescaled_annotation.setGroups(self.__annotation.getGroups())

            # Rescale and add the annotation items.
            #
            annotation_list = self.__annotation.getAnnotations()
            rescaled_annotation_list = []
            for annotation_item in annotation_list:
                # Rescale the coordinates.
                #
                annotation_item_coordinates = annotation_item.getCoordinates()
                rescaled_annotation_item_coordinates = [mir.Point(coordinate_item.getX() / factor, coordinate_item.getY() / factor) for coordinate_item in annotation_item_coordinates]

                # Rebuild the annotation item.
                #
                rescaled_annotation_item = mir.Annotation()
                rescaled_annotation_item.setName(annotation_item.getName())
                rescaled_annotation_item.setType(annotation_item.getType())
                rescaled_annotation_item.setGroup(rescaled_annotation.getGroup(annotation_item.getGroup().getName()))
                rescaled_annotation_item.setCoordinates(rescaled_annotation_item_coordinates)

                rescaled_annotation_list.append(rescaled_annotation_item)

            rescaled_annotation.setAnnotations(rescaled_annotation_list)

            # Return the rescaled annotation object.
            #
            return rescaled_annotation

        else:
            # No rescaling is required.
            #
            return self.__annotation

    @property
    def path(self):
        """
        Get the path of the opened annotation.

        Returns:
            str: Path of the opened annotation.
        """

        return self.__path

    @property
    def spacing(self):
        """
        Get the pixel spacing of the annotation.

        Returns:
            float: Pixel spacing.
        """

        return self.__spacing

    @property
    def groups(self):
        """
        Get the available annotation groups.

        Returns:
            list: Available annotation groups.
        """

        return list(self.__groups.keys())

    @property
    def counts(self):
        """
        Get the number of annotations in each annotation group.

        Returns:
            dict: Annotation count per group.
        """

        return self.__groups

    def convert(self, image_path, shape, pixel_spacing, label_map=None, conversion_order=None, strict=True):
        """
        Convert the annotations to a mask multi-resolution image file with the given shape pixel spacing.

        Args:
            image_path (str): Target mask image file path.
            shape (tuple): Required shape of the image.
            pixel_spacing (float, None): Pixel spacing of the mask (micrometer).
            label_map (dict): Mapping of the annotation groups to label values. The values must be in the [0, 255] range.
            conversion_order (list): Annotation group conversion order.
            strict (bool): Stop on unknown annotation groups.

        Raises:
            InvalidImageShapeError: Invalid image shape configuration.
            InvalidPixelSpacingError: Invalid pixel spacing configuration.
            AnnotationGroupsLabelMapMismatchError: The label map contains non-existing group name.
            AnnotationGroupsConversionOrderMismatchError: The conversion order list contains non-existing group name.
            LabelMapConversionOrderMismatchError: The label map does not match the conversion order list.
        """

        # Check the shape and the spacing.
        #
        if len(shape) != 2 or shape[0] <= 0 or shape[1] <= 0:
            raise dptimageerrors.InvalidImageShapeError(self.__path, shape)

        if pixel_spacing is not None and pixel_spacing <= 0.0:
            raise dptimageerrors.InvalidPixelSpacingError(self.__path, pixel_spacing)

        # Check if all the group names are valid in the label map and the conversion order.
        #
        if label_map is not None:
            if strict and any(group_name not in label_map for group_name in self.__groups):
                raise dptimageerrors.AnnotationGroupsLabelMapMismatchError(self.__path, list(self.__groups), label_map)

        if conversion_order is not None:
            if strict and any(group_name not in conversion_order for group_name in self.__groups):
                raise dptimageerrors.AnnotationGroupsConversionOrderMismatchError(self.__path, list(self.__groups), conversion_order)

        # Calculate output settings.
        #
        output_dimensions = tuple(reversed(shape))
        output_spacing = (pixel_spacing, pixel_spacing)
        conversion_label_map = dict(label_map) if label_map is not None else {group_name: 1 for group_name in self.__groups}
        conversion_order_list = list(conversion_order) if conversion_order is not None else list(conversion_label_map.keys())

        # Check the conversion order list against the label map.
        #
        if set(conversion_label_map.keys()) != set(conversion_order_list):
            raise dptimageerrors.LabelMapConversionOrderMismatchError(self.__path, list(label_map.keys()), conversion_order)

        # Rescale the annotations if necessary.
        #
        annotation_to_scale = self.__rescaledannotations(pixel_spacing=pixel_spacing)

        # Convert the annotation to mask image.
        #
        annotation_mask = mir.AnnotationToMask()
        annotation_mask.convert(annotation_to_scale, image_path, output_dimensions, output_spacing, conversion_label_map, conversion_order_list)
