"""
This module can convert ASAP annotation descriptor XMLs to mask files and clear the parts from a mask of an image
where the patch is completely empty and normalize the label values to the given interval.
"""

from ..errors import processingerrors as dptprocessingerrors
from ..image import imagereader as dptimagereader
from ..image import imagewriter as dptimagewriter
from ..image import annotationconverter as dptannotationconverter
from ..utils import imagefile as dptimagefile

import numpy as np
import datetime
import time
import os

#----------------------------------------------------------------------------------------------------

def _calculate_empty_map(image, tile_size, processing_level):
    """
    Calculates the map of non-empty tiles. If a tile contains non-zero value the value of its index at the result array is True, otherwise it is False.

    Args:
        image (dptimagereader.ImageReader): Image object to check.
        tile_size (int): Tile size.
        processing_level (int): Processing level of the image.

    Returns:
        (np.ndarray): Array of booleans indicating non completely empty patches at the given level of the input image.
    """

    # Calculate the size of the array. One bool for each image_size*image_size sized patch.
    #
    flag_array_shape = (image.shapes[processing_level][0] // tile_size + (1 if image.shapes[processing_level][0] % tile_size else 0),
                        image.shapes[processing_level][1] // tile_size + (1 if image.shapes[processing_level][1] % tile_size else 0))

    flag_array = np.zeros(shape=flag_array_shape, dtype=np.bool)

    # Go through the patches and mark non empty ones.
    #
    for row in range(0, flag_array_shape[0]):
        row_index = row * tile_size
        for col in range(0, flag_array_shape[1]):
            col_index = col * tile_size

            image_patch = image.read(spacing=image.spacings[processing_level], row=row_index, col=col_index, height=tile_size, width=tile_size)

            flag_array[row, col] = not np.all(np.logical_or(np.equal(image_patch, 0), np.equal(image_patch, 255)))

    # Return the per tile flag array.
    #
    return flag_array

#----------------------------------------------------------------------------------------------------

def _map_and_clear_image(mask_image, output_path, tile_size, label_map, empty_label, empty_map):
    """
    Map the label values and clear the marked empty parts from the given mask image.

    Args:
        mask_image (dptimagereader.ImageReader): Mask to process.
        output_path (str): Output mask path.
        tile_size (int): Processing tile size.
        label_map (dict): Label mapping.
        empty_label (int): Empty patch label value.
        empty_map (np.ndarray, None): Map of empty patches.

    Raises:
        DigitalPathologyConfigError: Configuration errors.
        DigitalPathologyImageError: Image errors.
    """

    # Configure the image writer.
    #
    image_writer = dptimagewriter.ImageWriter(image_path=output_path,
                                              shape=mask_image.shapes[0],
                                              spacing=mask_image.spacings[0],
                                              dtype=mask_image.dtype,
                                              coding=mask_image.coding,
                                              compression=None,
                                              interpolation=None,
                                              tile_size=tile_size,
                                              jpeg_quality=None,
                                              empty_value=empty_label,
                                              skip_empty=None)

    # Go through the patches and clear the empty ones.
    #
    if empty_map is None:
        for row in range(0, mask_image.shapes[0][0], tile_size):
            for col in range(0, mask_image.shapes[0][1], tile_size):
                image_patch = mask_image.read(spacing=mask_image.spacings[0], row=row, col=col, height=tile_size, width=tile_size)

                normalized_patch = image_patch.copy()

                for key, value in label_map.items():
                    normalized_patch[image_patch == key] = value

                image_writer.write(tile=normalized_patch, row=row, col=col)
    else:
        for row in range(0, empty_map.shape[0]):
            row_index = row * tile_size
            for col in range(0, empty_map.shape[1]):
                col_index = col * tile_size

                if empty_map[row, col]:
                    image_patch = mask_image.read(spacing=mask_image.spacings[0], row=row_index, col=col_index, height=tile_size, width=tile_size)

                    normalized_patch = image_patch.copy()

                    for key, value in label_map.items():
                        normalized_patch[image_patch == key] = value

                    image_writer.write(tile=normalized_patch, row=row_index, col=col_index)
                else:
                    image_writer.write(tile=None, row=row_index, col=col_index)

    # Finalize the output image.
    #
    image_writer.close()

#----------------------------------------------------------------------------------------------------

def create_annotation_mask(image,
                           annotation,
                           label_map,
                           conversion_order,
                           conversion_spacing,
                           spacing_tolerance,
                           output_path,
                           strict,
                           accept_all_empty,
                           copy_annotation_path=None,
                           work_output_path=None,
                           keep_copies=False,
                           overwrite=True,
                           logger=None):
    """
    Create mask files from annotation files.

    Args:
        image (dptimagereader.ImageReader, str): Image object or path.
        annotation (dptannotationconverter.AnnotationConverter, str): Annotation converter object or annotation file path.
        label_map (dict): Annotation group to label value map.
        conversion_order (list): Annotation group conversion order.
        conversion_spacing (float, None): Conversion pixel spacing of the input image (micrometer). The annotation is converted on the lowest level of the image if not specified.
        spacing_tolerance (float): Pixel spacing tolerance (percentage).
        output_path (str): Result mask file path.
        strict (bool): Stop on unknown annotation groups.
        accept_all_empty (bool): Accept if all the selected groups are empty.
        copy_annotation_path (str, None): Copy annotation path. If not None, the source annotation is copied to this path before processing. Not used if the annotation is a converter object.
        work_output_path (str, None): Work output path. If not None, the result mask image created here and copied to the target output path after processing.
        keep_copies (bool): Keep the copied annotation files.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Raises:
        EmptyAnnotationGroupsError: All the selected annotation group are empty and empty groups are not allowed.
        UnknownAnnotationGroupsError: There are unknown annotation groups present and strict mode is enabled.

        DigitalPathologyConfigError: Configuration errors.
        DigitalPathologyImageError: Image errors.
    """

    # Start time measurement.
    #
    start_time = time.time()

    # Construct source annotation paths and direct output paths.
    #
    if isinstance(annotation, dptannotationconverter.AnnotationConverter):
        input_annotation_path = annotation.path
    else:
        input_annotation_path = copy_annotation_path if copy_annotation_path else annotation

    output_conversion_path = work_output_path if work_output_path else output_path

    # Log parameters.
    #
    if logger:
        logger.info('Converting annotation to mask...')
        logger.info('Annotation: {path}'.format(path=annotation.path if isinstance(annotation, dptannotationconverter.AnnotationConverter) else annotation))
        logger.info('Image: {path}'.format(path=image.path if isinstance(image, dptimagereader.ImageReader) else image))
        logger.info('Mask: {path}'.format(path=output_path))
        logger.info('Label map: {map}'.format(map=label_map))
        logger.info('Conversion order: {order}'.format(order=conversion_order))
        logger.info('Conversion spacing: {spacing} um'.format(spacing=conversion_spacing))

    # Check if target already exits.
    #
    if not os.path.isfile(output_path) or overwrite:
        # Copy the source annotation file intermediate location for processing.
        #
        if not isinstance(annotation, dptannotationconverter.AnnotationConverter) and copy_annotation_path:
            dptimagefile.copy_image(source_path=annotation, target_path=copy_annotation_path, overwrite=overwrite, logger=logger)

        # Create work directory structure if necessary.
        #
        if work_output_path:
            os.makedirs(os.path.dirname(work_output_path), exist_ok=True)

        # Open image.
        #
        template_image = image if isinstance(image, dptimagereader.ImageReader) else dptimagereader.ImageReader(image_path=image, spacing_tolerance=spacing_tolerance, input_channels=None)
        used_conversion_spacing = conversion_spacing if conversion_spacing is not None else template_image.spacings[0]
        conversion_level = template_image.level(spacing=used_conversion_spacing)

        # Load the annotations from file.
        #
        annotation_conv = annotation if isinstance(annotation, dptannotationconverter.AnnotationConverter) else dptannotationconverter.AnnotationConverter(annotation_path=input_annotation_path,
                                                                                                                                                           pixel_spacing=template_image.spacings[0])

        # Check if there is at least one annotation.
        #
        if not accept_all_empty:
            total_point_count = 0
            for group_name in label_map:
                total_point_count += annotation_conv.counts.get(group_name, 0)

            if total_point_count == 0:
                raise dptprocessingerrors.EmptyAnnotationGroupsError(list(label_map.keys()))

        # Check if there is an unknown group name.
        #
        if strict:
            unknown_group_names = set(annotation_conv.groups) - set(label_map.keys())
            if unknown_group_names:
                # Raise error on unknown group.
                #
                raise dptprocessingerrors.UnknownAnnotationGroupsError(unknown_group_names)

        if logger:
            logger.info('Converting: {path}'.format(path=annotation_conv.path))

        # Convert the annotations to mask.
        #
        annotation_conv.convert(image_path=output_conversion_path,
                                shape=template_image.shapes[conversion_level],
                                pixel_spacing=template_image.spacings[conversion_level],
                                label_map=label_map,
                                conversion_order=conversion_order,
                                strict=strict)

        # Close the template image.
        #
        if not isinstance(image, dptimagereader.ImageReader):
            template_image.close()

        # Copy the result to the target location.
        #
        if work_output_path:
            if keep_copies:
                dptimagefile.copy_image(source_path=work_output_path, target_path=output_path, overwrite=overwrite, logger=logger)
            else:
                dptimagefile.move_image(source_path=work_output_path, target_path=output_path, overwrite=overwrite, logger=logger)

        # Remove duplicated annotation files.
        #
        if not keep_copies:
            if not isinstance(annotation, dptannotationconverter.AnnotationConverter) and copy_annotation_path and copy_annotation_path != annotation:
                dptimagefile.remove_image(image_path=copy_annotation_path, ignore_errors=True, logger=logger)

        # Log execution time.
        #
        execution_time = time.time() - start_time
        if logger:
            logger.debug('Done in {delta}'.format(delta=datetime.timedelta(seconds=execution_time)))

    else:
        if logger:
            logger.info('Skipping, target file already exits: {path}'.format(path=output_path))

#----------------------------------------------------------------------------------------------------

def create_annotation_mask_batch(job_list, label_map, conversion_order, conversion_spacing, spacing_tolerance, strict, accept_all_empty, keep_copies=False, overwrite=True, logger=None):
    """
    Go through all the listed jobs and create mask a file from each annotation file.

    Args:
        job_list (list): List of job (source image path, source annotation path, copy annotation path, work mask path, target mask path) quintets.
        label_map (dict): Annotation group to label value map.
        conversion_order (list): Annotation group conversion order.
        conversion_spacing (float, None): Conversion pixel spacing of the input image (micrometer). The annotation is converted on the lowest level of the image if not specified.
        spacing_tolerance (float): Pixel spacing tolerance (percentage).
        strict (bool): Stop on unknown annotation groups.
        accept_all_empty (bool): Accept if all the selected groups are empty.
        keep_copies (bool): Keep the copied annotation files.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Returns:
        (list, list): List of successfully processed items and list of failed items.
    """

    # Start time measurement.
    #
    start_time = time.time()

    # Log batch processing.
    #
    if logger:
        logger.info('Converting annotations to masks in batch mode...')
        logger.info('Job count: {count}'.format(count=len(job_list)))

    # Go through the job list.
    #
    failed_collection = []
    successful_collection = []
    for job_index in range(len(job_list)):
        # Get one job item.
        #
        source_image_path, source_annotation_path, copy_annotation_path, work_output_path, target_output_path = job_list[job_index]

        try:
            # Print data paths if the mode is not a single file mode.
            #
            if logger:
                logger.info('Processing [{index}/{count}]: {path}'.format(index=job_index+1, count=len(job_list), path=source_annotation_path))

            # Convert the annotation.
            #
            create_annotation_mask(image=source_image_path,
                                   annotation=source_annotation_path,
                                   label_map=label_map,
                                   conversion_order=conversion_order,
                                   conversion_spacing=conversion_spacing,
                                   spacing_tolerance=spacing_tolerance,
                                   output_path=target_output_path,
                                   strict=strict,
                                   accept_all_empty=accept_all_empty,
                                   copy_annotation_path=copy_annotation_path,
                                   work_output_path=work_output_path,
                                   keep_copies=keep_copies,
                                   overwrite=overwrite,
                                   logger=logger)

        except Exception as exception:
            if logger:
                logger.error('Error: {exception}'.format(exception=exception))

            # Add case to the error collection.
            #
            failed_collection.append(target_output_path)

        else:
            # Save the annotation to the list of successful conversions.
            #
            successful_collection.append(target_output_path)

    # Log execution time.
    #
    execution_time = time.time() - start_time
    if logger:
        logger.debug('Annotation conversion batch done in {delta}'.format(delta=datetime.timedelta(seconds=execution_time)))

    # Return a list of successful conversions.
    #
    return successful_collection, failed_collection

#----------------------------------------------------------------------------------------------------

def normalize_mask_image(mask, tile, output_path, label_map, empty_label, copy_mask_path=None, copy_tile_path=None, work_output_path=None, keep_copies=False, overwrite=True, logger=None):
    """
    Normalize mask image. First identify and fill the patches on the mask label with the empty value where the corresponding image is empty, then rearrange
    the label values in the mask according to the label value mapping.

    Args:
        mask (dptimagereader.ImageReader, str): Mask image object or path.
        tile (dptimagereader.ImageReader, str, None): Input image object or path. The image that will be used for empty part recognition.
        output_path (str): Output image path. The zoomed image.
        label_map (dict): Label mapping.
        empty_label (int): Empty patch label value.
        copy_mask_path (str, None): Copy mask path. If not None, the source mask image is copied to this path before processing. Not used if the mask is an image object.
        copy_tile_path (str, None): Copy tile path. If not None, the source tile image is copied to this path before processing. Not used if the tile is an image object.
        work_output_path (str, None): Work output path. If not None, the result mask image created here and copied to the target output path after processing.
        keep_copies (bool): Keep the copied mask and tile image files.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Raises:
        NoMatchingLevelInTileImageError: Cannot find matching level for mask in tile image.

        DigitalPathologyConfigError: Configuration errors.
        DigitalPathologyImageError: Image errors.
    """

    # Start time measurement.
    #
    start_time = time.time()

    # Construct source image paths and direct output paths.
    #
    if isinstance(mask, dptimagereader.ImageReader):
        input_mask_path = mask.path
    else:
        input_mask_path = copy_mask_path if copy_mask_path else mask

    if isinstance(tile, dptimagereader.ImageReader):
        input_tile_path = tile.path
    else:
        input_tile_path = copy_tile_path if copy_tile_path else tile

    output_normalization_path = work_output_path if work_output_path else output_path

    # Log parameters.
    #
    if logger:
        logger.info('Normalizing mask image...')
        logger.info('Mask: {path}'.format(path=mask.path if isinstance(mask, dptimagereader.ImageReader) else mask))
        logger.info('Template: {path}'.format(path=tile.path if isinstance(tile, dptimagereader.ImageReader) else tile))
        logger.info('Normalized: {path}'.format(path=output_path))
        logger.info('Label map: {map}'.format(map=label_map))
        logger.info('Empty label: {label}'.format(label=empty_label))

    # Check if the target image already exits.
    #
    if not os.path.isfile(output_path) or overwrite:
        # Copy the source mask and tile images to the intermediate locations for processing.
        #
        if not isinstance(mask, dptimagereader.ImageReader) and copy_mask_path:
            dptimagefile.copy_image(source_path=mask, target_path=copy_mask_path, overwrite=overwrite, logger=logger)

        if not isinstance(tile, dptimagereader.ImageReader) and copy_tile_path:
            dptimagefile.copy_image(source_path=tile, target_path=copy_tile_path, overwrite=overwrite, logger=logger)

        # Create work directory structure if necessary.
        #
        if work_output_path:
            os.makedirs(os.path.dirname(work_output_path), exist_ok=True)

        # Open images.
        #
        if tile is not None:
            tile_image = tile if isinstance(tile, dptimagereader.ImageReader) else dptimagereader.ImageReader(image_path=input_tile_path, spacing_tolerance=0.25, input_channels=None)
        else:
            tile_image = None

        mask_image = mask if isinstance(mask, dptimagereader.ImageReader) else dptimagereader.ImageReader(image_path=input_mask_path, spacing_tolerance=0.25, input_channels=None)

        if tile_image:
            if any(level_spacing is None for level_spacing in tile_image.spacings):
                tile_image.correct(spacing=1.0, level=0)

        if any(level_spacing is None for level_spacing in mask_image.spacings):
            mask_image.correct(spacing=1.0, level=0)

        # Basic configuration.
        #
        tile_size = 512  # Patch size for image reading and writing.

        # Calculate empty map.
        #
        if tile_image:
            # Find matching level with the tile image.
            #
            mask_dimensions = mask_image.shapes[0]
            level = None
            for level_index in range(len(tile_image.shapes)):
                if mask_dimensions == tile_image.shapes[level_index]:
                    level = level_index
                    break

            if level is None:
                raise dptprocessingerrors.NoMatchingLevelInTileImageError(tile_image.path, mask_image.path)

            if logger:
                logger.info('Identifying empty tiles: {path}'.format(path=tile_image.path))

            empty_map = _calculate_empty_map(image=tile_image, tile_size=tile_size, processing_level=level)
        else:
            empty_map = None

        # Clean empty parts.
        #
        if logger:
            logger.info('Normalizing mask: {path}'.format(path=output_normalization_path))

        _map_and_clear_image(mask_image=mask_image, output_path=output_normalization_path, tile_size=tile_size, label_map=label_map, empty_label=empty_label, empty_map=empty_map)

        # Close images.
        #
        if not isinstance(mask, dptimagereader.ImageReader):
            mask_image.close()

        if tile_image:
            if not isinstance(tile, dptimagereader.ImageReader):
                tile_image.close()

        # Copy the result to the target location.
        #
        if work_output_path:
            if keep_copies:
                dptimagefile.copy_image(source_path=work_output_path, target_path=output_path, overwrite=overwrite, logger=logger)
            else:
                dptimagefile.move_image(source_path=work_output_path, target_path=output_path, overwrite=overwrite, logger=logger)

        # Remove duplicated annotation files.
        #
        if not keep_copies:
            if not isinstance(mask, dptimagereader.ImageReader) and copy_mask_path and copy_mask_path != mask:
                dptimagefile.remove_image(image_path=copy_mask_path, ignore_errors=True, logger=logger)

            if not isinstance(tile, dptimagereader.ImageReader) and copy_tile_path and copy_tile_path != tile:
                dptimagefile.remove_image(image_path=copy_tile_path, ignore_errors=True, logger=logger)

        # Log execution time.
        #
        execution_time = time.time() - start_time
        if logger:
            logger.debug('Done in {delta}'.format(delta=datetime.timedelta(seconds=execution_time)))

    else:
        if logger:
            logger.info('Skipping, target file already exits: {path}'.format(path=output_path))

#----------------------------------------------------------------------------------------------------

def normalize_mask_image_batch(job_list, label_map, empty_label, keep_copies=False, overwrite=True, logger=None):
    """
    Normalize mask images. Tile path will be used to identify empty tiles.

    Args:
        job_list (list): List of job (source mask path, source tile path, copy mask path, copy tile path, work output path, target output path) sextets.
        label_map (dict): Label mapping.
        empty_label (int): Empty patch label value.
        overwrite (bool): If true existing targets will be overwritten.
        keep_copies (bool): Keep the copied annotation files.
        logger (logging.Logger, None): Logger object.

    Returns:
        (list, list): List of successfully processed items and list of failed items.
    """

    # Start time measurement.
    #
    start_time = time.time()

    # Log batch processing.
    #
    if logger:
        logger.info('Normalizing mask images in batch mode...')
        logger.info('Job count: {count}'.format(count=len(job_list)))

    # Go through the job list.
    #
    failed_collection = []
    successful_collection = []
    for job_index in range(len(job_list)):
        # Get one job item.
        #
        source_mask_path, source_tile_path, copy_mask_path, copy_tile_path, work_output_path, target_output_path = job_list[job_index]

        try:
            # Print data paths if the mode is not a single file mode.
            #
            if logger:
                logger.info('Processing [{index}/{count}]: {path}'.format(index=job_index+1, count=len(job_list), path=source_mask_path))

            # Normalize the mask image.
            #
            normalize_mask_image(mask=source_mask_path,
                                 tile=source_tile_path,
                                 output_path=target_output_path,
                                 label_map=label_map,
                                 empty_label=empty_label,
                                 copy_mask_path=copy_mask_path,
                                 copy_tile_path=copy_tile_path,
                                 work_output_path=work_output_path,
                                 keep_copies=keep_copies,
                                 overwrite=overwrite,
                                 logger=logger)

        except Exception as exception:
            if logger:
                logger.error('Error: {exception}'.format(exception=exception))

            # Add case to the error collection.
            #
            failed_collection.append(target_output_path)

        else:
            # Save the mask to the list of successful conversions.
            #
            successful_collection.append(target_output_path)

    # Log execution time.
    #
    execution_time = time.time() - start_time
    if logger:
        logger.debug('Mask image normalization batch done in {delta}'.format(delta=datetime.timedelta(seconds=execution_time)))

    # Return a list of successful normalizations.
    #
    return successful_collection, failed_collection
