"""
This module can evaluates the result of a tissue-background segmentation algorithm against a reference.
"""

import digitalpathology.image.imagereader as dptimagereader
import digitalpathology.processing.comparison as dptcomparison
import digitalpathology.processing.inference as dptinference
import digitalpathology.processing.plainimage as dptplainimage
import digitalpathology.processing.regions as dptregions
import digitalpathology.processing.traditional as dpttraditional
import digitalpathology.utils.imagefile as dptimagefile
import digitalpathology.utils.loggers as dptloggers


import pandas as ps
import numpy as np
import scipy.ndimage
import datetime
import time
import os
import sys

#----------------------------------------------------------------------------------------------------

def create_directory_structure(filtering_jobs, segmentation_jobs, comparison_jobs, preview_jobs, evaluation_jobs, logger=None):
    """
    Create target and work directory structure.

    Args:
        filtering_jobs (list): Reference filtering job tuples.
        segmentation_jobs (list): Segmentation job tuples.
        comparison_jobs (list): Comparison job tuples.
        preview_jobs (list): Preview job tuples.
        evaluation_jobs (list): Evaluation job tuples.
        logger (logging.Logger, None): Logger object.
    """

    # Start time measurement.
    #
    start_time = time.time()

    if logger:
        logger.info('Creating directory structure...')

    directories_to_create = set()

    # Collect directories for reference filtering.
    #
    for filtering_job_item in filtering_jobs:
        _, copy_mask_path, work_result_path, target_result_path = filtering_job_item

        if copy_mask_path:
            directories_to_create.add(os.path.dirname(copy_mask_path))

        if work_result_path:
            directories_to_create.add(os.path.dirname(work_result_path))

        if target_result_path:
            directories_to_create.add(os.path.dirname(target_result_path))

    # Collect the directories for segmentation.
    #
    for segmentation_job_item in segmentation_jobs:
        _, _, copy_image_path, copy_mask_path, work_result_path, work_interval_path, target_result_path, target_interval_path = segmentation_job_item

        if copy_image_path:
            directories_to_create.add(os.path.dirname(copy_image_path))

        if copy_mask_path:
            directories_to_create.add(os.path.dirname(copy_mask_path))

        if work_result_path:
            directories_to_create.add(os.path.dirname(work_result_path))

        if work_interval_path:
            directories_to_create.add(os.path.dirname(work_interval_path))

        if target_result_path:
            directories_to_create.add(os.path.dirname(target_result_path))

        if target_interval_path:
            directories_to_create.add(os.path.dirname(target_interval_path))

    # Collect the directories for comparison.
    #
    for comparison_job_item in comparison_jobs:
        _, _, copy_reference_path, copy_template_path, work_comparison_path, target_comparison_path = comparison_job_item

        if copy_reference_path:
            directories_to_create.add(os.path.dirname(copy_reference_path))

        if copy_template_path:
            directories_to_create.add(os.path.dirname(copy_template_path))

        if work_comparison_path:
            directories_to_create.add(os.path.dirname(work_comparison_path))

        if target_comparison_path:
            directories_to_create.add(os.path.dirname(target_comparison_path))

    # Collect the directories for preview.
    #
    for preview_job_item in preview_jobs:
        _, _, copy_image_path, copy_mask_path, work_preview_path, target_preview_path = preview_job_item

        if copy_image_path:
            directories_to_create.add(os.path.dirname(copy_image_path))

        if copy_mask_path:
            directories_to_create.add(os.path.dirname(copy_mask_path))

        if work_preview_path:
            directories_to_create.add(os.path.dirname(work_preview_path))

        if target_preview_path:
            directories_to_create.add(os.path.dirname(target_preview_path))

    # Collect the directories for evaluation.
    #
    for evaluation_job_item in evaluation_jobs:
        _, _, _, _, copy_mask_path, copy_reference_path, copy_template_path, copy_label_path = evaluation_job_item

        if copy_mask_path:
            directories_to_create.add(os.path.dirname(copy_mask_path))

        if copy_reference_path:
            directories_to_create.add(os.path.dirname(copy_reference_path))

        if copy_template_path:
            directories_to_create.add(os.path.dirname(copy_template_path))

        if copy_label_path:
            directories_to_create.add(os.path.dirname(copy_label_path))

    # Create collected directories.
    #
    for directory_path in directories_to_create:
        if logger:
            logger.info('Mkdir: {path}'.format(path=directory_path))

        os.makedirs(directory_path, exist_ok=True)

    # Log execution time.
    #
    exec_time = time.time() - start_time
    if logger:
        logger.debug('Directory structure setup done in {delta}'.format(delta=datetime.timedelta(seconds=exec_time)))

#----------------------------------------------------------------------------------------------------

def calculate_experiment_summary(experiment_table):
    """
    Calculate experiment summary: mean, standard deviation, minimal and maximal values of Dice score, sensitivity and false positive counts.

    Args:
        experiment_table:

    Returns:
        dict: Experiment summary.
    """

    dice_scores = experiment_table['dice score'].values
    sensitivities = experiment_table['sensitivity'].values
    false_positives = experiment_table['false positives'].values
    false_positive_ratios = experiment_table['false positive ratio'].values

    return {'dice score mean': np.mean(dice_scores),
            'dice score std': np.std(dice_scores),
            'dice score min': np.min(dice_scores),
            'dice score max': np.max(dice_scores),
            'sensitivity mean': np.mean(sensitivities),
            'sensitivity std': np.std(sensitivities),
            'sensitivity min': np.min(sensitivities),
            'sensitivity max': np.max(sensitivities),
            'false positives mean': np.mean(false_positives),
            'false positives std': np.std(false_positives),
            'false positives min': np.min(false_positives),
            'false positives max': np.max(false_positives),
            'false positive ratios mean': np.mean(false_positive_ratios),
            'false positive ratios std': np.std(false_positive_ratios),
            'false positive ratios min': np.min(false_positive_ratios),
            'false positive ratios max': np.max(false_positive_ratios)}

#----------------------------------------------------------------------------------------------------

def filter_reference_batch(job_list,
                           diagonal_threshold,
                           reference_label,
                           background_label,
                           pixel_spacing,
                           spacing_tolerance,
                           keep_copies,
                           overwrite,
                           logger=None):
    """
    Compare segmentation against the reference masks.

    Args:
        job_list (list): List of job (source mask path, copy mask path, work result path, target result path) quartets. Source and target paths are mandatory. If the copy path is set, the
            source item is copied to the copy path before processing.
        diagonal_threshold (float): Region diagonal minimum (micrometer), for removing smaller than threshold sized regions from the reference.
        reference_label (int): Reference label.
        background_label (int): Replacement label label.
        pixel_spacing (float): Processing pixel spacing (micrometer).
        spacing_tolerance (float): Pixel spacing tolerance (percentage).
        keep_copies (bool): Keep the copied image files.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Returns:
        (list, list): List of successfully processed items and list of failed items.
    """

    # Start time measurement.
    #
    start_time = time.time()

    if logger:
        logger.info('Filtering reference...')

    successful_collection, failed_collection = dptregions.filter_regions_batch(job_list=job_list,
                                                                               diagonal_threshold=diagonal_threshold,
                                                                               full_connectivity=True,
                                                                               pixel_spacing=pixel_spacing,
                                                                               spacing_tolerance=spacing_tolerance,
                                                                               foreground_labels=[reference_label],
                                                                               background_label=background_label,
                                                                               keep_copies=keep_copies,
                                                                               overwrite=overwrite,
                                                                               logger=logger)

    # Log execution time.
    #
    exec_time = time.time() - start_time
    if logger:
        logger.debug('Reference filtering done in {delta}'.format(delta=datetime.timedelta(seconds=exec_time)))

    # Return a list of successful and failed comparisons.
    #
    return successful_collection, failed_collection

#----------------------------------------------------------------------------------------------------

def segment_image_batch(job_list,
                        method_name,
                        model_path,
                        gray_threshold,
                        patch_size,
                        template_label,
                        segmentation_pixel_spacing,
                        output_pixel_spacing,
                        spacing_tolerance,
                        channels_first,
                        input_channels,
                        confidence_threshold,
                        region_diagonal_threshold,
                        hole_diagonal_threshold,
                        overwrite,
                        logger):
    """
    Apply segmentation using CNN of one of the traditional methods: Otsu or FESI.

    Args:
        job_list (list): Segmentation job tuples.
        method_name (str): Segmentation method name: 'cnn', 'otsu', 'fesi', or 'threshold'.
        model_path (str): Model file path.
        gray_threshold (int, None): Graysclae threshold for 'threshold' method.
        patch_size (int): Inference patch size.
        template_label (int): Template label.
        segmentation_pixel_spacing (float): Pixel spacing for segmentation (micrometers).
        output_pixel_spacing (float): Output pixel spacing of the segmentation (micrometers). Zooming is applied on the network output if necessary.
        spacing_tolerance (float): Pixel spacing tolerance (percentage).
        channels_first (bool): If true, the network expect patches shaped BCHW, otherwise BHWC.
        input_channels (list): Input channel indices to use for inference.
        confidence_threshold (float): Confidence threshold to apply on the inference images for binary masks.
        region_diagonal_threshold (float): Region diagonal minimum (micrometer), for removing smaller than threshold sized regions from the output.
        hole_diagonal_threshold (float): Hole diagonal minimum (micrometer), for filling smaller than threshold sized tissue holes from the output.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Returns:
        (list, list): List of successfully processed items and list of failed items.
    """

    # Start time measurement.
    #
    start_time = time.time()

    if logger:
        logger.info('Segmenting images...')

    # Execute segmentation.
    #
    if method_name == 'cnn':
        # Calculate segmentation with network inference.
        #
        successful_collection, failed_collection = dptinference.apply_network_batch(job_list=job_list,
                                                                                    model_path=model_path,
                                                                                    patch_size=patch_size,
                                                                                    output_class=template_label,
                                                                                    number_of_classes=-1,
                                                                                    normalizer='rgb_to_0-1',
                                                                                    normalizer_source_range=None,
                                                                                    normalizer_target_range=None,
                                                                                    soft_mode=True,
                                                                                    input_spacing=segmentation_pixel_spacing,
                                                                                    output_spacing=output_pixel_spacing,
                                                                                    spacing_tolerance=spacing_tolerance,
                                                                                    unrestrict_network=True,
                                                                                    channels_first=channels_first,
                                                                                    input_channels=input_channels,
                                                                                    padding_mode='reflect',
                                                                                    padding_constant=255,
                                                                                    confidence=confidence_threshold,
                                                                                    minimum_region_diagonal=region_diagonal_threshold,
                                                                                    minimum_hole_diagonal=hole_diagonal_threshold,
                                                                                    full_connectivity=True,
                                                                                    quantize=False,
                                                                                    interpolation_order=1,
                                                                                    keep_intermediates=False,
                                                                                    keep_copies=True,
                                                                                    single_mode=False,
                                                                                    overwrite=overwrite,
                                                                                    logger=logger)

    elif method_name in ['otsu', 'fesi', 'threshold']:
        # Rearrange the job items: get rid of the unnecessary items for the traditional methods.
        #
        traditional_job_list = [(source_path, copy_path, work_path, target_path) for source_path, _, copy_path, _, work_path, _, target_path, _ in job_list]

        # Calculate segmentation with Otsu's threshold or with the FESI algorithm or with the simple thresholding.
        #
        successful_collection, failed_collection = dpttraditional.calculate_tissue_mask_batch(job_list=traditional_job_list,
                                                                                              pixel_spacing=output_pixel_spacing,
                                                                                              spacing_tolerance=spacing_tolerance,
                                                                                              method=method_name,
                                                                                              gray_threshold=gray_threshold,
                                                                                              minimum_region_diagonal=region_diagonal_threshold,
                                                                                              minimum_hole_diagonal=hole_diagonal_threshold,
                                                                                              keep_copies=True,
                                                                                              overwrite=overwrite,
                                                                                              logger=logger)
    else:
        # Unknown method.
        #
        if logger:
            logger.error('Error: Unknown segmentation methods: {method}'.format(method=method_name))

        # Add all segmentation target result paths to the failed items list.
        #
        successful_collection = []
        failed_collection = [job_item[6] for job_item in job_list]

    # Log execution time.
    #
    exec_time = time.time() - start_time
    if logger:
        logger.debug('Image segmentation done in {delta}'.format(delta=datetime.timedelta(seconds=exec_time)))

    # Return a list of successful and failed comparisons.
    #
    return successful_collection, failed_collection

#----------------------------------------------------------------------------------------------------

def compare_segmentation_batch(job_list,
                               reference_label,
                               template_label,
                               pixel_spacing,
                               spacing_tolerance,
                               true_detection_threshold,
                               false_detection_threshold,
                               keep_copies,
                               overwrite,
                               logger=None):
    """
    Compare segmentation against the reference masks.

    Args:
        job_list (list): List of job (source reference path, source template path, copy reference path, copy template path, work comparison path, target comparison path) sextets. Source and target
            paths are mandatory. If the copy paths are set, the source items are copied to the copy paths before processing.
        reference_label (int): Reference label.
        template_label (int): Template label.
        pixel_spacing (float): Processing pixel spacing (micrometer).
        spacing_tolerance (float): Pixel spacing tolerance (percentage).
        true_detection_threshold (float): A reference region is detected (TP) if larger or equal than threshold ratio of it overlaps with the template, otherwise missed (FN).
        false_detection_threshold (float): A template region is a detection if larger or equal than threshold area ratio of it overlaps with the reference, otherwise false detection (FP).
        keep_copies (bool): Keep the copied image files.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Returns:
        (list, list): List of successfully processed items and list of failed items.
    """

    # Start time measurement.
    #
    start_time = time.time()

    if logger:
        logger.info('Comparing segmentation to the reference...')

    # Go through the job list.
    #
    failed_collection = []
    successful_collection = []
    for job_index in range(len(job_list)):
        # Get one job item.
        #
        source_reference_path, source_template_path, copy_reference_path, copy_template_path, work_comparison_path, target_comparison_path = job_list[job_index]

        # Construct direct input/output paths.
        #
        input_reference_path = copy_reference_path if copy_reference_path else source_reference_path
        input_template_path = copy_template_path if copy_template_path else source_template_path
        output_comparison_path = work_comparison_path if work_comparison_path else target_comparison_path

        # Print progress.
        #
        if logger:
            logger.info('Processing [{index}/{count}]: {path}'.format(index=job_index + 1, count=len(job_list), path=target_comparison_path))

        try:
            # Check if the target image already exits.
            #
            if not os.path.isfile(target_comparison_path) or overwrite:
                # Copy the reference and template masks to the intermediate locations for processing.
                #
                if copy_reference_path:
                    dptimagefile.copy_image(source_path=source_reference_path, target_path=copy_reference_path, overwrite=overwrite, logger=logger)

                if copy_template_path:
                    dptimagefile.copy_image(source_path=source_template_path, target_path=copy_template_path, overwrite=overwrite, logger=logger)

                # Create work directory structure if necessary.
                #
                if work_comparison_path:
                    os.makedirs(os.path.dirname(work_comparison_path), exist_ok=True)

                # Compare the masks.
                #
                dptcomparison.image_comparison(reference=input_reference_path,
                                               reference_label=reference_label,
                                               template=input_template_path,
                                               template_label=template_label,
                                               pixel_spacing=pixel_spacing,
                                               spacing_tolerance=spacing_tolerance,
                                               true_detection_threshold=true_detection_threshold,
                                               false_detection_threshold=false_detection_threshold,
                                               full_connectivity=True,
                                               output_path=output_comparison_path,
                                               overwrite=overwrite,
                                               detected_positive_label=1,
                                               detected_negative_label=2,
                                               missed_positive_label=3,
                                               missed_negative_label=4,
                                               false_positive_label=5,
                                               false_negative_label=6,
                                               logger=logger)

                # Copy the result to the target location.
                #
                if work_comparison_path:
                    if keep_copies:
                        dptimagefile.copy_image(source_path=work_comparison_path, target_path=target_comparison_path, overwrite=overwrite, logger=logger)
                    else:
                        dptimagefile.move_image(source_path=work_comparison_path, target_path=target_comparison_path, overwrite=overwrite, logger=logger)

            else:
                if logger:
                    logger.info('Skipping, target file already exits: {path}'.format(path=target_comparison_path))

        except Exception as exception:
            if logger:
                logger.error('Error: {exception}'.format(exception=exception))

            # Add case to the error collection.
            #
            failed_collection.append(target_comparison_path)

        else:
            # Save the result to the list of successful comparisons.
            #
            successful_collection.append(target_comparison_path)

    # Log execution time.
    #
    exec_time = time.time() - start_time
    if logger:
        logger.debug('Segmentation comparision done in {delta}'.format(delta=datetime.timedelta(seconds=exec_time)))

    # Return a list of successful and failed comparisons.
    #
    return successful_collection, failed_collection

#----------------------------------------------------------------------------------------------------

def calculate_preview_batch(job_list, pixel_spacing, spacing_tolerance, keep_copies, overwrite, logger=None):
    """
    Calculate segmentation preview thumbnails.

    Args:
        job_list (list): List of job (source image path, source mask path, copy image path, copy mask path, work preview path, target preview path) sextets.
        pixel_spacing (float): Processing pixel spacing (micrometer).
        spacing_tolerance (float): Pixel spacing tolerance (percentage).
        keep_copies (bool): Keep the copied image files.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Returns:
        (list, list): List of successfully processed items and list of failed items.
    """

    # Start time measurement.
    #
    start_time = time.time()

    if logger:
        logger.info('Calculatig previews...')

    # Define custom palette for preview.
    #
    preview_palette = [[0, 0, 0], [0, 255, 0], [0, 0, 255], [255, 255, 0], [0, 255, 255], [255, 0, 0], [128, 0, 255]]

    successful_collection, failed_collection = dptplainimage.calculate_preview_batch(job_list=job_list,
                                                                                     level=None,
                                                                                     pixel_spacing=pixel_spacing,
                                                                                     spacing_tolerance=spacing_tolerance,
                                                                                     alpha=0.5,
                                                                                     palette=preview_palette,
                                                                                     keep_copies=keep_copies,
                                                                                     overwrite=overwrite,
                                                                                     logger=logger)

    # Log execution time.
    #
    exec_time = time.time() - start_time
    if logger:
        logger.debug('Preview calculation done in {delta}'.format(delta=datetime.timedelta(seconds=exec_time)))

    # Return a list of successful and failed calculations.
    #
    return successful_collection, failed_collection

#----------------------------------------------------------------------------------------------------

def false_positive_ratio_image(image_path, pixel_spacing, spacing_tolerance, label_value, logger):
    """
    Calculate the false positive area ratio of the image.

    Args:
        image_path (str): Mask image file path.
        pixel_spacing (float): Processing pixel spacing (micrometer).
        spacing_tolerance (float): Pixel spacing tolerance (percentage).
        label_value (int): Label value of the false positive regions.
        logger (logging.Logger, None): Logger object.

    Returns:
        float: False positive area ratio.
    """

    # Start time measurement.
    #
    start_time = time.time()

    # Log parameters.
    #
    if logger:
        logger.info('Calculating false positive area ratio...')
        logger.info('Image: {path}'.format(path=image_path))
        logger.info('Label: {label}'.format(label=label_value))

    # Load image into memory.
    #
    image = dptimagereader.ImageReader(image_path=image_path, spacing_tolerance=spacing_tolerance, input_channels=None)
    image_level = image.level(spacing=pixel_spacing)
    image_content = image.content(spacing=pixel_spacing)

    # Count the number of labels.
    #
    label_count = np.count_nonzero(np.equal(image_content, label_value))
    label_ratio = label_count / (image.shapes[image_level][0] * image.shapes[image_level][1])

    # Log the result.
    #
    if logger:
        logger.info('False positive ratio: {ratio}'.format(ratio=label_ratio))

    # Close the image.
    #
    image.close()

    # Log execution time.
    #
    exec_time = time.time() - start_time
    if logger:
        logger.debug('Done in {delta}'.format(delta=datetime.timedelta(seconds=exec_time)))

    # Return the calculated ratio.
    #
    return label_ratio

#----------------------------------------------------------------------------------------------------

def count_removed_reference_image(mask_path, reference_path, pixel_spacing, spacing_tolerance, logger):
    """
    Count the number of removed regions.

    Args:
        mask_path:
        reference_path:
        pixel_spacing (float): Processing pixel spacing (micrometer).
        spacing_tolerance (float): Pixel spacing tolerance (percentage).
        logger (logging.Logger, None): Logger object.

    Returns:
        int: Number of removed false positive regions.
    """

    # Start time measurement.
    #
    start_time = time.time()

    # Log parameters.
    #
    if logger:
        logger.info('Counting removed reference regions...')
        logger.info('Mask: {path}'.format(path=mask_path))
        logger.info('Filtered: {path}'.format(path=reference_path))

    # Load images into memory.
    #
    mask = dptimagereader.ImageReader(image_path=mask_path, spacing_tolerance=spacing_tolerance, input_channels=None)
    mask_content = mask.content(spacing=pixel_spacing).squeeze()

    reference = dptimagereader.ImageReader(image_path=reference_path, spacing_tolerance=spacing_tolerance, input_channels=None)
    reference_content = reference.content(spacing=pixel_spacing).squeeze()

    difference_array = np.not_equal(mask_content, reference_content)

    connectivity_structure = np.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], dtype=np.bool)
    _, difference_region_count = scipy.ndimage.measurements.label(input=difference_array, structure=connectivity_structure)

    # Log the result.
    #
    if logger:
        logger.info('Removed reference regions: {count}'.format(count=difference_region_count))

    # Close the images.
    #
    reference.close()
    mask.close()

    # Log execution time.
    #
    exec_time = time.time() - start_time
    if logger:
        logger.debug('Done in {delta}'.format(delta=datetime.timedelta(seconds=exec_time)))

    # Return the region count.
    #
    return difference_region_count

#----------------------------------------------------------------------------------------------------

def evaluate_segmentation_batch(job_list,
                                output_path,
                                reference_label,
                                template_label,
                                true_detection_threshold,
                                false_detection_threshold,
                                pixel_spacing,
                                spacing_tolerance,
                                keep_copies,
                                overwrite,
                                logger):
    """
    Evaluate the segmentation by calculating Dice score, Jaccard index, sensitivity, true positives, false negatives, and false positives.

    Args:
        job_list (list): List of job (source mask path, source reference path, source template path, source label path, copy mask path, copy reference path, copy template path, copy label path)
            octets. Source paths are mandatory. If the copy paths are set, the source items are copied to the copy paths before processing.
        output_path (str): Output table path.
        reference_label (int): Reference label.
        template_label (int): Template label.
        true_detection_threshold (float): A reference region is detected (TP) if larger or equal than threshold ratio of it overlaps with the template, otherwise missed (FN).
        false_detection_threshold (float): A template region is a detection if larger or equal than threshold area ratio of it overlaps with the reference, otherwise false detection (FP).
        pixel_spacing (float): Processing pixel spacing (micrometer).
        spacing_tolerance (float): Pixel spacing tolerance (percentage).
        keep_copies (bool): Keep the copied image files.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Returns:
        (list, list): List of successfully processed items and list of failed items.
    """

    # Start time measurement.
    #
    start_time = time.time()

    if logger:
        logger.info('Evaluating segmentation...')

    # Prepare output collections.
    #
    failed_collection = []
    successful_collection = []

    if not os.path.isfile(output_path) or overwrite:
        # Create pandas data frame.
        #
        output_columns = ['name', 'dice score', 'jaccard index', 'sensitivity', 'true positives', 'false negatives', 'false positives', 'false positive ratio', 'removed reference']
        output_df = ps.DataFrame(columns=output_columns)

        # Go through the job list.
        #
        for job_index in range(len(job_list)):
            # Get one job item.
            #
            source_mask_path, source_reference_path, source_template_path, source_label_path, copy_mask_path, copy_reference_path, copy_template_path, copy_label_path = job_list[job_index]

            # Construct direct input/output paths.
            #
            input_mask_path = copy_mask_path if copy_mask_path else source_mask_path
            input_reference_path = copy_reference_path if copy_reference_path else source_reference_path
            input_template_path = copy_template_path if copy_template_path else source_template_path
            input_label_path = copy_label_path if copy_label_path else source_label_path

            try:
                # Print data paths if the mode is not a single file mode.
                #
                if logger:
                    logger.info('Processing [{index}/{count}]: {path}'.format(index=job_index+1, count=len(job_list), path=source_reference_path))

                if copy_mask_path:
                    dptimagefile.copy_image(source_path=source_mask_path, target_path=copy_mask_path, overwrite=overwrite, logger=logger)

                if copy_reference_path:
                    dptimagefile.copy_image(source_path=source_reference_path, target_path=copy_reference_path, overwrite=overwrite, logger=logger)

                if copy_template_path:
                    dptimagefile.copy_image(source_path=source_template_path, target_path=copy_template_path, overwrite=overwrite, logger=logger)

                if copy_label_path:
                    dptimagefile.copy_image(source_path=source_label_path, target_path=copy_label_path, overwrite=overwrite, logger=logger)

                # Calculate Dice score and Jaccard index.
                #
                dice_score, jaccard_index = dptcomparison.image_similarity(reference=input_reference_path,
                                                                           reference_label=reference_label,
                                                                           template=input_template_path,
                                                                           template_label=template_label,
                                                                           pixel_spacing=pixel_spacing,
                                                                           spacing_tolerance=spacing_tolerance,
                                                                           logger=logger)

                # Calculate image statistics: true positives, false negatives, and false positives.
                #
                true_positive_count, false_negative_count, false_positive_count = dptcomparison.image_statistics(reference=input_reference_path,
                                                                                                                 reference_label=reference_label,
                                                                                                                 template=input_template_path,
                                                                                                                 template_label=template_label,
                                                                                                                 pixel_spacing=pixel_spacing,
                                                                                                                 spacing_tolerance=spacing_tolerance,
                                                                                                                 true_detection_threshold=true_detection_threshold,
                                                                                                                 false_detection_threshold=false_detection_threshold,
                                                                                                                 full_connectivity=True,
                                                                                                                 logger=logger)

                # Calculate false positive area ratio.
                #
                false_positive_ratio = false_positive_ratio_image(image_path=input_label_path, pixel_spacing=pixel_spacing, spacing_tolerance=spacing_tolerance, label_value=6, logger=logger)

                # Count the number of removed reference regions.
                #
                removed_reference = count_removed_reference_image(mask_path=input_mask_path,
                                                                  reference_path=input_reference_path,
                                                                  pixel_spacing=pixel_spacing,
                                                                  spacing_tolerance=spacing_tolerance,
                                                                  logger=logger)

                # Calculate sensitivity.
                #
                sensitivity = true_positive_count / (true_positive_count + false_negative_count)

                # Add result to the data frame.
                #
                output_row = {'name': os.path.basename(source_reference_path),
                              'dice score': dice_score,
                              'jaccard index': jaccard_index,
                              'sensitivity': sensitivity,
                              'true positives': true_positive_count,
                              'false negatives': false_negative_count,
                              'false positives': false_positive_count,
                              'false positive ratio': false_positive_ratio,
                              'removed reference': removed_reference}

                output_df = output_df.append(ps.DataFrame(data=output_row, index=[0]), ignore_index=True)

                # Remove duplicated image files.
                #
                if not keep_copies:
                    if copy_mask_path and copy_mask_path != source_mask_path:
                        dptimagefile.remove_image(image_path=copy_mask_path, ignore_errors=True, logger=logger)

                    if copy_reference_path and copy_reference_path != source_reference_path:
                        dptimagefile.remove_image(image_path=copy_reference_path, ignore_errors=True, logger=logger)

                    if copy_template_path and copy_template_path != source_template_path:
                        dptimagefile.remove_image(image_path=copy_template_path, ignore_errors=True, logger=logger)

                    if copy_label_path and copy_label_path != source_label_path:
                        dptimagefile.remove_image(image_path=copy_label_path, ignore_errors=True, logger=logger)

            except Exception as exception:
                if logger:
                    logger.error('Error: {exception}'.format(exception=exception))

                # Add case to the error collection.
                #
                failed_collection.append(source_reference_path)

            else:
                # Save the result to the list of successful zooms.
                #
                successful_collection.append(source_reference_path)

        # Sort the table by image file name.
        #
        output_df = output_df.sort_values(by=['name'])

        # Save the result table.
        #
        output_df.to_csv(path_or_buf=output_path, columns=output_columns, index=False)

        # Log summary.
        #
        experiment_summary = calculate_experiment_summary(experiment_table=output_df)

        if logger:
            logger.info('Dice score mean: {mean}'.format(mean=experiment_summary['dice score mean']))
            logger.info('Dice score std: {std}'.format(std=experiment_summary['dice score std']))
            logger.info('Sensitivity mean: {mean}'.format(mean=experiment_summary['sensitivity mean']))
            logger.info('Sensitivity std: {std}'.format(std=experiment_summary['sensitivity std']))
            logger.info('False positives mean: {mean}'.format(mean=experiment_summary['false positives mean']))
            logger.info('False positives std: {std}'.format(std=experiment_summary['false positives std']))
            logger.info('False positive ratios mean: {mean}'.format(mean=experiment_summary['false positive ratios mean']))
            logger.info('False positive ratios std: {std}'.format(std=experiment_summary['false positive ratios std']))

    else:
        if logger:
            logger.info('Skipping: target file already exits')

    # Log execution time.
    #
    exec_time = time.time() - start_time
    if logger:
        logger.debug('Segmentation evaluation done in {delta}'.format(delta=datetime.timedelta(seconds=exec_time)))

    # Return a list of successful zooms.
    #
    return successful_collection, failed_collection

#----------------------------------------------------------------------------------------------------

def clean_up_intermediate_files(filtering_jobs, segmentation_jobs, comparison_jobs, preview_jobs, evaluation_jobs, keep_filtering, keep_segmentation, keep_comparison, logger=None):
    """
    Clean up the intermediate files after processing.

    Args:
        filtering_jobs (list): Reference filtering job tuples.
        segmentation_jobs (list): Segmentation job tuples.
        comparison_jobs (list): Comparison job tuples.
        preview_jobs (list): Preview job tuples.
        evaluation_jobs (list): Evaluation job tuples.
        keep_filtering (bool): Keep filtered reference files.
        keep_segmentation (bool): Keep segmentation output files.
        keep_comparison (bool): Keep comparison output files.
        logger (logging.Logger, None): Logger object.
    """

    # Start time measurement.
    #
    start_time = time.time()

    if logger:
        logger.info('Cleaning up reference filtering files...')

    for filtering_job_item in filtering_jobs:
        # Get one job item.
        #
        source_mask_path, copy_mask_path, work_result_path, target_result_path = filtering_job_item

        # Remove mask file copy.
        #
        if copy_mask_path and copy_mask_path != source_mask_path:
            dptimagefile.remove_image(image_path=copy_mask_path, ignore_errors=True, logger=logger)

        # Remove the output.
        #
        if work_result_path and work_result_path != target_result_path:
            dptimagefile.remove_image(image_path=work_result_path, ignore_errors=True, logger=logger)

        # Remove filtering target file if configured.
        #
        if not keep_filtering:
            if target_result_path:
                dptimagefile.remove_image(image_path=target_result_path, ignore_errors=True, logger=logger)

    if logger:
        logger.info('Cleaning up segmentation files...')

    # Clean up the leftover files after segmentation.
    #
    for segmentation_job_item in segmentation_jobs:
        # Get one job item.
        #
        source_image_path, source_mask_path, copy_image_path, copy_mask_path, work_result_path, work_interval_path, target_result_path, target_interval_path = segmentation_job_item

        # Remove the image and mask file copies.
        #
        if copy_image_path and copy_image_path != source_image_path:
            dptimagefile.remove_image(image_path=copy_image_path, ignore_errors=True, logger=logger)

        if copy_mask_path and copy_mask_path != source_mask_path:
            dptimagefile.remove_image(image_path=copy_mask_path, ignore_errors=True, logger=logger)

        # Remove the outputs.
        #
        if work_result_path and work_result_path != target_result_path:
            dptimagefile.remove_image(image_path=work_result_path, ignore_errors=True, logger=logger)

        if work_interval_path and work_interval_path != target_interval_path:
            dptimagefile.remove_image(image_path=work_interval_path, ignore_errors=True, logger=logger)

        # Remove segmentation target files if configured.
        #
        if not keep_segmentation:
            if target_result_path:
                dptimagefile.remove_image(image_path=target_result_path, ignore_errors=True, logger=logger)

            if target_interval_path:
                dptimagefile.remove_image(image_path=target_interval_path, ignore_errors=True, logger=logger)

    if logger:
        logger.info('Cleaning up comparison files...')

    # Clean up the leftover files after comparison.
    #
    for comparison_job_item in comparison_jobs:
        # Get one job item.
        #
        source_reference_path, source_template_path, copy_reference_path, copy_template_path, work_comparison_path, target_comparison_path = comparison_job_item

        # Clean up the source images.
        #
        if copy_reference_path and copy_reference_path != source_reference_path:
            dptimagefile.remove_image(image_path=copy_reference_path, ignore_errors=True, logger=logger)

        if copy_template_path and copy_template_path != source_template_path:
            dptimagefile.remove_image(image_path=copy_template_path, ignore_errors=True, logger=logger)

        # Clean up the result comparison files.
        #
        if work_comparison_path and work_comparison_path != target_comparison_path:
            dptimagefile.remove_image(image_path=work_comparison_path, ignore_errors=True, logger=logger)

        # Remove comparison target file if configured.
        #
        if not keep_comparison:
            if target_comparison_path:
                dptimagefile.remove_image(image_path=target_comparison_path, ignore_errors=True, logger=logger)

    if logger:
        logger.info('Cleaning up preview files...')

    # Clean up the leftover files after preview.
    #
    for preview_job_item in preview_jobs:
        # Get one job item.
        #
        source_image_path, source_mask_path, copy_image_path, copy_mask_path, work_preview_path, target_preview_path = preview_job_item

        # Clean up the source images.
        #
        if copy_image_path and copy_image_path != source_image_path:
            dptimagefile.remove_image(image_path=copy_image_path, ignore_errors=True, logger=logger)

        if copy_mask_path and copy_mask_path != source_mask_path:
            dptimagefile.remove_image(image_path=copy_mask_path, ignore_errors=True, logger=logger)

        # Clean up the result preview files.
        #
        if work_preview_path and work_preview_path != target_preview_path:
            dptimagefile.remove_image(image_path=work_preview_path, ignore_errors=True, logger=logger)

    if logger:
        logger.info('Cleaning up evaluation files...')

    # Clean up the leftover files after evaluation.
    #
    for evaluation_job_item in evaluation_jobs:
        # Get one job item.
        #
        source_mask_path, source_reference_path, source_template_path, source_label_path, copy_mask_path, copy_reference_path, copy_template_path, copy_label_path = evaluation_job_item

        # Clean up the source images.
        #
        if copy_mask_path and copy_mask_path != source_mask_path:
            dptimagefile.remove_image(image_path=copy_mask_path, ignore_errors=True, logger=logger)

        if copy_reference_path and copy_reference_path != source_reference_path:
            dptimagefile.remove_image(image_path=copy_reference_path, ignore_errors=True, logger=logger)

        if copy_template_path and copy_template_path != source_template_path:
            dptimagefile.remove_image(image_path=copy_template_path, ignore_errors=True, logger=logger)

        if copy_label_path and copy_label_path != source_label_path:
            dptimagefile.remove_image(image_path=copy_label_path, ignore_errors=True, logger=logger)

    # Log execution time.
    #
    exec_time = time.time() - start_time
    if logger:
        logger.debug('Clean up done in {delta}'.format(delta=datetime.timedelta(seconds=exec_time)))

#----------------------------------------------------------------------------------------------------

def evaluate_model(filtering_jobs,
                   segmentation_jobs,
                   comparison_jobs,
                   preview_jobs,
                   evaluation_jobs,
                   segmentation_pixel_spacing,
                   evaluation_pixel_spacing,
                   preview_pixel_spacing,
                   output_table_path,
                   log_path,
                   segmentation_method_name,
                   model_path,
                   gray_threshold,
                   patch_size,
                   reference_label,
                   template_label,
                   true_detection_threshold,
                   false_detection_threshold,
                   confidence_threshold,
                   reference_diagonal_threshold,
                   region_diagonal_threshold,
                   hole_diagonal_threshold,
                   input_channels,
                   spacing_tolerance,
                   channels_first,
                   keep_filtering,
                   keep_segmentation,
                   keep_comparison,
                   overwrite):
    """
    Evaluate the network model, by using it to inference, and evaluating the generated masks or use the traditional methods for segmentation.

    Args:
        filtering_jobs (list): Reference filtering job tuples.
        segmentation_jobs (list): Segmentation job tuples.
        comparison_jobs (list): Comparison job tuples.
        preview_jobs (list): Preview job tuples.
        evaluation_jobs (list): Evaluation job tuples.
        segmentation_pixel_spacing (float): Pixel spacing for segmentation (micrometers).
        evaluation_pixel_spacing (float): Pixel spacing for evaluation (micrometers).
        preview_pixel_spacing (float, None): Pixel spacing for preview (micrometers). If not configured, previews are not produced.
        output_table_path (str): Output summary table path expression.
        log_path (str, None): Log file path expression.
        segmentation_method_name (str): Segmentation method name: 'cnn', 'otsu', 'fesi', or 'threshold'.
        model_path (str): Model file path.
        gray_threshold (int, None): Graysclae threshold for 'threshold' method.
        patch_size (int): Inference patch size.
        reference_label (int): Reference label.
        template_label (int): Template label.
        true_detection_threshold (float): A reference region is detected (TP) if larger or equal than threshold ratio of it overlaps with the template, otherwise missed (FN).
        false_detection_threshold (float): A template region is a detection if larger or equal than threshold area ratio of it overlaps with the reference, otherwise false detection (FP).
        confidence_threshold (float): Confidence threshold to apply on the inference images for binary masks.
        reference_diagonal_threshold (float): Region diagonal minimum (micrometer), for removing smaller than threshold sized regions from the reference.
        region_diagonal_threshold (float): Region diagonal minimum (micrometer), for removing smaller than threshold sized regions from the output.
        hole_diagonal_threshold (float): Hole diagonal minimum (micrometer), for filling smaller than threshold sized tissue holes from the output.
        input_channels (list): Input channel indices to use for inference.
        spacing_tolerance (float): Pixel spacing tolerance (percentage).
        channels_first (bool): If true, the network expect patches shaped BCHW, otherwise BHWC.
        keep_filtering (bool): Keep filtered reference files.
        keep_segmentation (bool): Keep segmentation output files.
        keep_comparison (bool): Keep comparison output files.
        overwrite (bool): If true existing targets will be overwritten.

    Returns:
        (list, list): List of successfully processed items and list of failed items.
    """

    # Start time measurement.
    #
    start_time = time.time()

    # Initialize logging.
    #
    os.makedirs(os.path.dirname(log_path), exist_ok=True)

    logger_name = '{namespace}.{function}'.format(namespace=__name__, function=evaluate_model.__name__)
    logger = dptloggers.filelogger(name=logger_name, log_path=log_path, debug=True)

    logger.info('Evaluating method...')

    # Log all arguments.
    #
    logger.debug('Arguments/Filtering job count: {count}'.format(count=len(filtering_jobs)))
    logger.debug('Arguments/Segmentation job count: {count}'.format(count=len(segmentation_jobs)))
    logger.debug('Arguments/Comparison job count: {count}'.format(count=len(comparison_jobs)))
    logger.debug('Arguments/Preview job count: {count}'.format(count=len(preview_jobs)))
    logger.debug('Arguments/Evaluation job count: {count}'.format(count=len(evaluation_jobs)))
    logger.debug('Arguments/Segmentation pixel spacing: {spacing} um'.format(spacing=segmentation_pixel_spacing))
    logger.debug('Arguments/Evaluation pixel spacing: {spacing} um'.format(spacing=evaluation_pixel_spacing))
    logger.debug('Arguments/Preview pixel spacing: {spacing} um'.format(spacing=preview_pixel_spacing))
    logger.debug('Arguments/Output table path: {path}'.format(path=output_table_path))
    logger.debug('Arguments/Method name: {method}'.format(method=segmentation_method_name))
    logger.debug('Arguments/Model path: {path}'.format(path=model_path))
    logger.debug('Arguments/Gray threshold: {threshold}'.format(threshold=gray_threshold))
    logger.debug('Arguments/Patch size: {size}'.format(size=patch_size))
    logger.debug('Arguments/Reference label: {label}'.format(label=reference_label))
    logger.debug('Arguments/Template label: {label}'.format(label=template_label))
    logger.debug('Arguments/True detection threshold: {threshold}'.format(threshold=true_detection_threshold))
    logger.debug('Arguments/False detection threshold: {threshold}'.format(threshold=false_detection_threshold))
    logger.debug('Arguments/Confidence threshold: {threshold}'.format(threshold=confidence_threshold))
    logger.debug('Arguments/Reference region diagonal threshold: {threshold} um'.format(threshold=reference_diagonal_threshold))
    logger.debug('Arguments/Segmentation region diagonal threshold: {threshold} um'.format(threshold=region_diagonal_threshold))
    logger.debug('Arguments/Segmentation hole diagonal threshold: {threshold} um'.format(threshold=hole_diagonal_threshold))
    logger.debug('Arguments/Input channels: {channels}'.format(channels=input_channels))
    logger.debug('Arguments/Spacing tolerance: {tolerance}'.format(tolerance=spacing_tolerance))
    logger.debug('Arguments/Channels first: {flag}'.format(flag=channels_first))
    logger.debug('Arguments/Keep filtering: {flag}'.format(flag=keep_filtering))
    logger.debug('Arguments/Keep segmentation: {flag}'.format(flag=keep_segmentation))
    logger.debug('Arguments/Keep comparison: {flag}'.format(flag=keep_comparison))
    logger.debug('Arguments/Overwrite existing results: {flag}'.format(flag=overwrite))

    # Prepare result arrays.
    #
    successful_filtering_items = []
    successful_segmentation_items = []
    successful_preview_items = []
    successful_evaluation_items = []
    failed_filtering_items = []
    failed_segmentation_items = []
    failed_preview_items = []
    failed_evaluation_items = []

    # Check if the target image already exits.
    #
    if not os.path.isfile(output_table_path) or overwrite:
        # Create target directory structure.
        #
        create_directory_structure(filtering_jobs=filtering_jobs,
                                   segmentation_jobs=segmentation_jobs,
                                   comparison_jobs=comparison_jobs,
                                   preview_jobs=preview_jobs,
                                   evaluation_jobs=evaluation_jobs,
                                   logger=logger)

        # Filter out the smaller than threshold regions from the reference.
        #
        successful_filtering_items, failed_filtering_items = filter_reference_batch(job_list=filtering_jobs,
                                                                                    diagonal_threshold=reference_diagonal_threshold,
                                                                                    reference_label=reference_label,
                                                                                    background_label=0,
                                                                                    pixel_spacing=evaluation_pixel_spacing,
                                                                                    spacing_tolerance=spacing_tolerance,
                                                                                    keep_copies=True,
                                                                                    overwrite=overwrite,
                                                                                    logger=logger)

        # Use the network or the traditional method for segmentation and keep the copied image files.
        #
        successful_segmentation_items, failed_segmentation_items = segment_image_batch(job_list=segmentation_jobs,
                                                                                       method_name=segmentation_method_name,
                                                                                       model_path=model_path,
                                                                                       gray_threshold=gray_threshold,
                                                                                       patch_size=patch_size,
                                                                                       template_label=template_label,
                                                                                       segmentation_pixel_spacing=segmentation_pixel_spacing,
                                                                                       output_pixel_spacing=evaluation_pixel_spacing,
                                                                                       spacing_tolerance=spacing_tolerance,
                                                                                       channels_first=channels_first,
                                                                                       input_channels=input_channels,
                                                                                       confidence_threshold=confidence_threshold,
                                                                                       region_diagonal_threshold=region_diagonal_threshold,
                                                                                       hole_diagonal_threshold=hole_diagonal_threshold,
                                                                                       overwrite=overwrite,
                                                                                       logger=logger)

        # Compare the reference mask and the segmentation output.
        #
        successful_comparison_items, failed_comparison_items = compare_segmentation_batch(job_list=comparison_jobs,
                                                                                          reference_label=reference_label,
                                                                                          template_label=template_label,
                                                                                          pixel_spacing=evaluation_pixel_spacing,
                                                                                          spacing_tolerance=spacing_tolerance,
                                                                                          true_detection_threshold=true_detection_threshold,
                                                                                          false_detection_threshold=false_detection_threshold,
                                                                                          keep_copies=True,
                                                                                          overwrite=overwrite,
                                                                                          logger=logger)

        # Calculate previews of the segmentation.
        #
        successful_preview_items, failed_preview_items = calculate_preview_batch(job_list=preview_jobs,
                                                                                 pixel_spacing=preview_pixel_spacing,
                                                                                 spacing_tolerance=spacing_tolerance,
                                                                                 keep_copies=True,
                                                                                 overwrite=overwrite,
                                                                                 logger=logger)

        # Evaluate the model.
        #
        successful_evaluation_items, failed_evaluation_items = evaluate_segmentation_batch(job_list=evaluation_jobs,
                                                                                           output_path=output_table_path,
                                                                                           reference_label=reference_label,
                                                                                           template_label=template_label,
                                                                                           true_detection_threshold=true_detection_threshold,
                                                                                           false_detection_threshold=false_detection_threshold,
                                                                                           pixel_spacing=evaluation_pixel_spacing,
                                                                                           spacing_tolerance=spacing_tolerance,
                                                                                           keep_copies=False,
                                                                                           overwrite=overwrite,
                                                                                           logger=logger)

        # Clean up.
        #
        clean_up_intermediate_files(filtering_jobs=filtering_jobs,
                                    segmentation_jobs=segmentation_jobs,
                                    comparison_jobs=comparison_jobs,
                                    preview_jobs=preview_jobs,
                                    evaluation_jobs=evaluation_jobs,
                                    keep_filtering=keep_filtering,
                                    keep_segmentation=keep_segmentation,
                                    keep_comparison=keep_comparison,
                                    logger=logger)

        # Log summary.
        #
        if logger:
            logger.info('Successful mask filterings: {count}/{total}'.format(count=len(successful_filtering_items), total=len(filtering_jobs)))
            logger.info('Successful image segmentations: {count}/{total}'.format(count=len(successful_segmentation_items), total=len(segmentation_jobs)))
            logger.info('Successful mask comparisons: {count}/{total}'.format(count=len(successful_comparison_items), total=len(comparison_jobs)))
            logger.info('Successful preview generations: {count}/{total}'.format(count=len(successful_preview_items), total=len(preview_jobs)))
            logger.info('Successful mask evaluations: {count}/{total}'.format(count=len(successful_evaluation_items), total=len(evaluation_jobs)))

        # Log execution time.
        #
        execution_time = time.time() - start_time
        if logger:
            logger.info('Complete evaluation done in {delta}'.format(delta=datetime.timedelta(seconds=execution_time)))

    else:
        if logger:
            logger.info('Skipping, target file already exits: {path}'.format(path=output_table_path))

    # Return a list of successful and failed cases.
    #
    successful_items = successful_filtering_items + successful_segmentation_items + successful_preview_items + successful_evaluation_items
    failed_items = failed_filtering_items + failed_segmentation_items + failed_preview_items + failed_evaluation_items

    return successful_items, failed_items
