"""
This module can load a network and apply classification on a whole slide image.
"""

from . import regions as dptregions
from . import threshold as dptthreshold
from . import zoom as dptzoom

from ..errors import processingerrors as dptprocessingerrors
from ..image import imagereader as dptimagereader
from ..models import modelbase as dptmodelbase
from ..normalizers import generalnormalizer as dtpgeneralnormalizer
from ..normalizers import passthroughnormalizer as dtppassthroughnormalizer
from ..normalizers import rgbnormalizer as dtprgbnormalizer
from ..normalizers import rgbtozeroonenormalizer as dtprgbtozeroonenormalizer
from ..utils import imagefile as dptimagefile
from ..utils import network as dptnetwork
from ..utils import serialize as dptserialize

import datetime
import time
import os
import numpy as np

#----------------------------------------------------------------------------------------------------

def _construct_intermediate_file_paths(output_path, apply_zoom, apply_threshold, apply_diagonal_filter, apply_hole_filling, prediction_suffix, zoom_suffix, threshold_suffix, filter_suffix):
    """
    Construct intermediate file paths for prediction and zooming.

    Args:
        output_path (str): Output image file path.
        apply_zoom (int): Zooming is required.
        apply_threshold (bool): Thresholding is required.
        apply_diagonal_filter (bool): Region filtering is required.
        apply_hole_filling (bool): Hole filling is required.
        prediction_suffix (str): Prediction file name suffix.
        zoom_suffix (str): Zoom output file name suffix.
        threshold_suffix (str): Thresholded file name suffix.
        filter_suffix (str): Filtered file name suffix.

    Returns:
        str, str, str, str, str: Intermediate file names for prediction, zoom, thresholded, filtered, and filled files
    """

    # Construct file paths for intermediate files.
    #
    output_path_parts = os.path.splitext(output_path)

    if not apply_zoom and not apply_threshold and not apply_diagonal_filter and not apply_hole_filling:
        prediction_path = output_path
    else:
        prediction_path = '{base}{suffix}{ext}'.format(base=output_path_parts[0], suffix=prediction_suffix, ext=output_path_parts[1])

    if apply_zoom:
        if not apply_threshold and not apply_diagonal_filter and not apply_hole_filling:
            zoomed_path = output_path
        else:
            zoomed_path = '{base}{suffix}{ext}'.format(base=output_path_parts[0], suffix=zoom_suffix, ext=output_path_parts[1])
    else:
        zoomed_path = prediction_path

    if apply_threshold:
        if not apply_diagonal_filter and not apply_hole_filling:
            thresholded_path = output_path
        else:
            thresholded_path = '{base}{suffix}{ext}'.format(base=output_path_parts[0], suffix=threshold_suffix, ext=output_path_parts[1])
    else:
        thresholded_path = zoomed_path

    if apply_diagonal_filter:
        if not apply_hole_filling:
            filtered_path = output_path
        else:
            filtered_path = '{base}{suffix}{ext}'.format(base=output_path_parts[0], suffix=filter_suffix, ext=output_path_parts[1])
    else:
        filtered_path = thresholded_path

    filled_path = output_path

    # Return the calculated file paths.
    #
    return prediction_path, zoomed_path, thresholded_path, filtered_path, filled_path

#----------------------------------------------------------------------------------------------------

def _load_network_model(model_path, required_shape, unrestrict_network, logger):
    """
    Load network model and instantiate it.

    Args:
        model_path (str): Path of the model file to load.
        required_shape (tuple): Processing patch shape.
        unrestrict_network (bool): Fix network for fixed in and output sizes.
        logger (logging.Logger, None): Logger object.

    Returns:
        digitalpathology.models.modelbase.ModelBase: Loaded network instance.

    Raises:
        MissingModelFileError: Missing model file.

        DigitalPathologyModelError: Model errors.
    """

    # Check if the model path is a valid file path.
    #
    if not os.path.isfile(path=model_path):
        raise dptprocessingerrors.MissingModelFileError(model_path)

    # Load the network data structure separately, in case the input shape needs to be unfixed.
    #
    if logger:
        logger.info('Loading network ...')

    network_data = dptserialize.load_object(path=model_path)

    # Return the prepared model.
    #
    network = dptmodelbase.ModelBase.instantiate(file=network_data)

    if unrestrict_network:
        # Check if the actual model can load the required input shape.
        #
        input_shape = network.model.input_shape
        output_shape = network.model.output_shape

        # Check if the input shape of the network is identical to the required shape. Or if the output of the network is 2 (N, classes). If it is more then 2,
        # we assume the predict function will return the correct output shape.
        #
        if not all(input_shape[index] is None or input_shape[index] == required_shape[index] for index in range(len(required_shape))) or len(output_shape) == 2:
            # Try to unfix the input shape and re-instantiate the network.
            #
            if logger:
                logger.info('Unfixing input shape ...')

            unfixed_network_data = dptnetwork.free_network_input_shape(network=network_data)
            network = dptmodelbase.ModelBase.instantiate(file=unfixed_network_data)

    return network

#----------------------------------------------------------------------------------------------------

def _find_mask_level(input_image, input_spacing, mask_image):
    """
    Find the mask level to load based on image pixel spacing.

    Args:
        input_image (dptimagereader.ImageReader): Input image object.
        input_spacing (float): Pixel spacing of the input image to use.
        mask_image (dptimagereader.ImageReader, None): Mask image object.

    Returns:
        int: Mask image level.
    """

    if mask_image is not None:
        # If the spacing information is missing from the mask file, assume that the spacing of the mask is exactly the same as of the image file where the shape of the two files first match.
        #
        if any(mask_spacing is None for mask_spacing in mask_image.spacings):
            for image_level in range(len(input_image.shapes)):
                if input_image.shapes[image_level] == mask_image.shapes[0]:
                    mask_image.correct(spacing=input_image.spacings[image_level], level=0)
                    break

        # Find the corresponding level int the mask image.
        #
        return mask_image.level(spacing=input_spacing)
    else:
        # There is no mask image.
        #
        return -1

#----------------------------------------------------------------------------------------------------

def _init_classifier(network_model, patch_size, padding_mode, padding_constant, output_class, soft_classification, quantize, class_count, channels_first):
    """
    Initialize image classifier.

    Args:
        network_model (digitalpathology.models.modelbase.ModelBase): Network instance.
        patch_size (int): Patch size to set for processing.
        padding_mode (str): Padding mode. Accepted modes are: 'constant' and 'reflect.
        padding_constant (int): Padding constant to use with 'constnat' padding mode.
        output_class (int): Output class for soft classification. If negative all class predictions will be written out in separate channels.
        soft_classification (bool): Soft classification flag. If true the output will be floating point from the [0.0, 1.0] interval, otherwise binary.
        quantize (bool): Quantization flag. If true and soft classification is selected the [0.0, 1.0] interval will be quantized to [0, 255].
        class_count (int): Number of output classes.
        channels_first (bool): If true then channels first will be used.

    Returns:
        WholeSlideImageClassifier: The configured image classifier object.
    """

    # Local import to prevent unnecessary GPU initialization.
    #
    import digitalpathology.classification.wholeslideimageclassifier as dptwholeslideimageclassifier

    # Configure classifier.
    #
    image_classifier = dptwholeslideimageclassifier.WholeSlideImageClassifier()
    image_classifier.model = network_model
    image_classifier.mode = 'fully_convolutional'
    image_classifier.tile_size = patch_size
    image_classifier.quantize = quantize
    image_classifier.soft = soft_classification
    image_classifier.verbose = True
    image_classifier.axis_description = 'cwh' if channels_first else 'whc'
    image_classifier.channels_in_result = class_count
    image_classifier.padding_mode = padding_mode
    image_classifier.padding_constant = padding_constant

    if 0 <= output_class:
        image_classifier.output_class = output_class

    return image_classifier

#----------------------------------------------------------------------------------------------------

def _init_normalizer(normalizer_id, source_range, target_range):
    """
    Initializes a normalizer object.

    Args:
        normalizer_id (str): Normalizer identifier.
        source_range (tuple, None): Normalizer source range in (min, max) format.
        target_range (tuple, None): Normalizer target range in (min, max) format.

    Returns:
        function: A preprocessing normalizer function for the classifier.

    Raises:
        UnknownImageNormalizationModeError: Unknown image normalization mode.
    """

    # Configure normalizer_id
    #
    if normalizer_id == 'general':
        validation_normalizer = dtpgeneralnormalizer.GeneralNormalizer(target_range=target_range, source_range=source_range)

    elif normalizer_id == 'rgb':
        validation_normalizer = dtprgbnormalizer.RgbNormalizer(target_range=target_range)

    elif normalizer_id == 'rgb_to_0-1':
        validation_normalizer = dtprgbtozeroonenormalizer.RgbToZeroOneNormalizer()

    elif normalizer_id == 'pass-through':
        validation_normalizer = dtppassthroughnormalizer.PassThroughNormalizer()

    else:
        # The given normalization type is unknown.
        #
        raise dptprocessingerrors.UnknownImageNormalizationModeError(normalizer_id)

    # Define preprocessing function for the classifier.
    #
    def normalizer_function(image):
        batch_dict = {0: {'patches': np.array([image]), 'labels': np.empty(0), 'weight': np.empty(0)}}
        return validation_normalizer.normalize(batch_dict)[0]['patches'][0]

    return normalizer_function

#----------------------------------------------------------------------------------------------------

def apply_network_batch(job_list,
                        model_path,
                        patch_size,
                        output_class,
                        number_of_classes,
                        normalizer,
                        normalizer_source_range,
                        normalizer_target_range,
                        soft_mode,
                        input_spacing,
                        output_spacing=None,
                        spacing_tolerance=0.25,
                        unrestrict_network=False,
                        channels_first=False,
                        input_channels=(0, 1, 2),
                        padding_mode='constant',
                        padding_constant=255,
                        confidence=0.5,
                        minimum_region_diagonal=0.0,
                        minimum_hole_diagonal=0.0,
                        full_connectivity=False,
                        quantize=True,
                        interpolation_order=0,
                        keep_intermediates=False,
                        keep_copies=False,
                        single_mode=False,
                        overwrite=True,
                        logger=None):
    """
    Apply network on batch of images.

    Args:
        job_list (list): List of job (source image path, source mask path, copy image path, copy mask path, work result path, work interval path, target result path, target interval path)
            octets where interval path is the quantization interval file path. If empty or None the interval will not be written out. Source image and target result paths are mandatory, the rest
            is optional. If copy paths are set, the images copied from the source paths to the copy paths before processing. If work paths are set, the output files are first written to the work
            paths then copied to the target path after processing.
        model_path (str): Path of the model file to load.
        patch_size (int): Patch size to set for processing.
        output_class (int): Output class for soft classification. If negative all class predictions will be written out in separate channels.
        number_of_classes (int): Number of output classes. If negative will try to get number of channels of network.
        normalizer (str): Normalizer identifier.
        normalizer_source_range (tuple, None): Normalizer source range in (min, max) format.
        normalizer_target_range (tuple, None): Normalizer target range in (min, max) format.
        soft_mode (bool): Soft classification. If set too true the output will be the softmax.
        input_spacing (float): Pixel spacing level of WSI processed by the classifier.
        output_spacing (float, None): Required output pixel spacing: level 0 pixel spacing of the output image will be the given pixel spacing level of the input image.
        spacing_tolerance (float): Pixel spacing tolerance (percentage).
        unrestrict_network (bool): If true the network will be fixed (delete fixed input_size and output_size).
        channels_first (bool): If true, the network expect patches shaped BCHW, otherwise BHWC.
        input_channels (list): List of input channel indices.
        padding_mode (str): Padding mode. Accepted modes are: 'constant' and 'reflect'.
        padding_constant (int): Padding constant to use with 'constnat' padding mode.
        confidence (float): Network confidence for thresholding the predictions. Thresholding is omitted for threshold values smaller or equal to 0.0.
        minimum_region_diagonal (float): Region diagonal minimum (micrometer), for removing smaller than threshold sized regions from the output. Executed only if the diagonal is larger than 0.0.
        minimum_hole_diagonal (float): Hole diagonal minimum (micrometer), for filling smaller than threshold sized holes from the output. Executed only if the diagonal is larger than 0.0.
        full_connectivity (bool): Connectivity matrix for region and hole filtering. If true edge and point neighbors will be used otherwise edge neighbors only.
        quantize (bool): Quantization flag. If true the network output is quantized to [0, 255] uint8 interval.
        interpolation_order (int): Interpolation order from the [0, 5] interval.
        keep_intermediates (bool): Keep intermediate files.
        keep_copies (bool): Keep the copied image files.
        single_mode (bool): If true the batch execution acts as if it was a single call: exceptions raised, no processing count logged.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Returns:
        (list, list): List of successfully processed items and list of failed items.

    Raises:
        MissingModelFileError: Missing model file.
        UnknownImageNormalizationModeError: Unknown image normalization mode.
        InvalidPixelSpacingValueError: Both image level and pixel spacing are None.
        InvalidInterpolationOrderError: Interpolation order out of [0, 5] bounds.
        AsymmetricZoomFactorError: Asymmetric zoom values.
        InvalidZoomFactorError: Non-positive zoom factor.

        DigitalPathologyConfigError: Configuration errors.
        DigitalPathologyImageError: Image errors.
        DigitalPathologyModelError: Model errors.
        DigitalPathologyProcessingError: Processing errors.
    """

    # Start time measurement.
    #
    batch_start_time = time.time()

    # Log batch processing.
    #
    if not single_mode:
        if logger:
            logger.info('Applying network in batch mode...')
            logger.info('Job count: {count}'.format(count=len(job_list)))

    try:
        # Calculate the required input shape for the network.
        #
        input_shape = (None, patch_size, patch_size, len(input_channels))

        # The model should be only loaded once:
        #
        network_model = _load_network_model(model_path=model_path, required_shape=input_shape, unrestrict_network=unrestrict_network, logger=logger)

    except Exception as exception:
        # Re-raise exception for the single file mode.
        #
        if single_mode:
            raise

        if logger:
            logger.error('Error: {exception}'.format(exception=exception))

        # Add all cases to the error collection and return: network inference is not possible without the loaded network.
        #
        return [], [target_result_path for _, _, _, _, _, _, target_result_path, _ in job_list]

    # Go through the job list.
    #
    failed_collection = []
    successful_collection = []
    for job_index in range(len(job_list)):
        # Get one job item.
        #
        source_image_path, source_mask_path, copy_image_path, copy_mask_path, work_result_path, work_interval_path, target_result_path, target_interval_path = job_list[job_index]

        # Construct direct input/output paths.
        #
        input_image_path = copy_image_path if copy_image_path else source_image_path
        input_mask_path = copy_mask_path if copy_mask_path else source_mask_path
        output_result_path = work_result_path if work_result_path else target_result_path
        output_interval_path = work_interval_path if work_interval_path else target_interval_path

        # Print data paths if the mode is not a single file mode.
        #
        if logger:
            if not single_mode:
                logger.info('Processing [{index}/{count}]: {path}'.format(index=job_index + 1, count=len(job_list), path=target_result_path))

            logger.info('Applying network...')
            logger.info('Model path: {path}'.format(path=model_path))
            logger.info('Input path: {path}'.format(path=source_image_path))
            logger.info('Mask path: {path}'.format(path=source_mask_path))
            logger.info('Result path: {path}'.format(path=target_result_path))
            logger.info('Input spacing: {spacing} um'.format(spacing=input_spacing))
            logger.info('Output class: {output} of {count} classes'.format(output=output_class, count=number_of_classes))
            logger.info('Normalizer: \'{normalizer}\''.format(normalizer=normalizer))
            logger.info('Soft mode: {mode}'.format(mode=soft_mode))

        try:
            # Check if the target image already exits.
            #
            if not os.path.isfile(target_result_path) or overwrite:
                # Progress reporting.
                #
                segmentation_start_time = time.time()

                # Copy the source image and mask to the intermediate locations for processing.
                #
                if copy_image_path:
                    dptimagefile.copy_image(source_path=source_image_path, target_path=copy_image_path, overwrite=overwrite, logger=logger)

                if copy_mask_path:
                    dptimagefile.copy_image(source_path=source_mask_path, target_path=copy_mask_path, overwrite=overwrite, logger=logger)

                # Open the input image.
                #
                input_image = dptimagereader.ImageReader(image_path=input_image_path, spacing_tolerance=spacing_tolerance, input_channels=input_channels)

                # Open the mask image.
                #
                mask_image = dptimagereader.ImageReader(image_path=input_mask_path, spacing_tolerance=spacing_tolerance, input_channels=None) if input_mask_path else None

                # Create work directory structure if necessary.
                #
                if work_result_path:
                    os.makedirs(os.path.dirname(work_result_path), exist_ok=True)

                if work_interval_path:
                    os.makedirs(os.path.dirname(work_interval_path), exist_ok=True)

                # Initialize image classifier object: it is always executed in soft mode and the output will be thresholded manually later.
                #
                image_classifier = _init_classifier(network_model=network_model,
                                                    patch_size=patch_size,
                                                    padding_mode=padding_mode,
                                                    padding_constant=padding_constant,
                                                    output_class=output_class,
                                                    soft_classification=soft_mode,
                                                    quantize=quantize,
                                                    channels_first=channels_first,
                                                    class_count=number_of_classes)

                # Determine processing steps.
                #
                apply_zoom = output_spacing is not None and input_image.level(spacing=output_spacing) != input_image.level(spacing=input_spacing)
                apply_threshold = soft_mode and 0.0 < confidence
                apply_diagonal_filter = 0.0 < minimum_region_diagonal
                apply_hole_filling = 0.0 < minimum_hole_diagonal

                # Construct file names.
                #
                prediction_path, zoomed_path, thresholded_path, filtered_path, filled_path = _construct_intermediate_file_paths(output_path=output_result_path,
                                                                                                                                apply_zoom=apply_zoom,
                                                                                                                                apply_threshold=apply_threshold,
                                                                                                                                apply_diagonal_filter=apply_diagonal_filter,
                                                                                                                                apply_hole_filling=apply_hole_filling,
                                                                                                                                prediction_suffix='_prediction',
                                                                                                                                zoom_suffix='_zoomed',
                                                                                                                                threshold_suffix='_thresholded',
                                                                                                                                filter_suffix='_filtered')

                # Configure inputs and outputs for the classifier.
                #
                input_level = input_image.level(spacing=input_spacing)
                output_level = input_image.level(spacing=output_spacing) if output_spacing is not None else input_level
                mask_mr_image = mask_image.image if mask_image else None
                mask_level = _find_mask_level(input_image=input_image, input_spacing=input_spacing, mask_image=mask_image)

                image_classifier.set_input(image=input_image.image, level=input_level, channels=list(input_channels), mask=mask_mr_image,  mask_level=mask_level)
                image_classifier.set_result_file_path(path=prediction_path)

                # Configure normalizer.
                #
                image_classifier.preprocess_function = _init_normalizer(normalizer_id=normalizer, source_range=normalizer_source_range, target_range=normalizer_target_range)

                # Classify image.
                #
                image_classifier.process()

                # Re-calculate confidence threshold based on quantization interval length.
                #
                if quantize:
                    quantization_interval = image_classifier.quantize_min_and_max
                    low_threshold = min(max(0, int(quantization_interval[0] + (quantization_interval[1] - quantization_interval[0]) * confidence)), 255)
                else:
                    quantization_interval = None
                    low_threshold = confidence

                # Write out the quantization interval file path.
                #
                if quantize and output_interval_path:
                    np.save(file=output_interval_path, arr=np.asarray(quantization_interval, dtype=np.float64), allow_pickle=True)

                # Measure inference time.
                #
                inference_exec_time = time.time() - segmentation_start_time
                if logger:
                    logger.debug('Done in {delta}'.format(delta=datetime.timedelta(seconds=inference_exec_time)))

                # Zoom image.
                #
                if apply_zoom:
                    # Check level difference: zoom in or zoom out = save at the given level.
                    #
                    if output_spacing < input_spacing:
                        dptzoom.zoom_image(image=prediction_path,
                                           output_path=zoomed_path,
                                           zoom=input_image.shapes[output_level],
                                           level=0,
                                           pixel_spacing=None,
                                           spacing_tolerance=spacing_tolerance,
                                           round_shape=True,
                                           interpolation_order=interpolation_order,
                                           jpeg_quality=None,
                                           overwrite=overwrite,
                                           logger=logger)
                    else:
                        dptzoom.save_image_at_level(image=prediction_path,
                                                    output_path=zoomed_path,
                                                    level=output_level,
                                                    pixel_spacing=None,
                                                    spacing_tolerance=spacing_tolerance,
                                                    jpeg_quality=None,
                                                    overwrite=overwrite,
                                                    logger=logger)

                # Threshold the prediction to get the classification.
                #
                if apply_threshold:
                    dptthreshold.low_threshold_image(image=zoomed_path,
                                                     output_path=thresholded_path,
                                                     low_threshold=low_threshold,
                                                     load_all=False,
                                                     overwrite=overwrite,
                                                     logger=logger)

                # Filter the regions according to their size.
                #
                if apply_diagonal_filter:
                    dptregions.filter_regions_image(image=thresholded_path,
                                                    output_path=filtered_path,
                                                    diagonal_threshold=minimum_region_diagonal,
                                                    full_connectivity=full_connectivity,
                                                    pixel_spacing=None,
                                                    spacing_tolerance=spacing_tolerance,
                                                    foreground_labels=None,
                                                    background_label=0,
                                                    copy_input_path=None,
                                                    work_output_path=None,
                                                    keep_copies=False,
                                                    overwrite=overwrite,
                                                    logger=logger)

                # Fill holes according to their size.
                #
                if apply_hole_filling:
                    dptregions.fill_holes_image(image=filtered_path,
                                                output_path=filled_path,
                                                diagonal_threshold=minimum_hole_diagonal,
                                                full_connectivity=full_connectivity,
                                                pixel_spacing=None,
                                                spacing_tolerance=spacing_tolerance,
                                                foreground_labels=None,
                                                fill_value=1,
                                                copy_input_path=None,
                                                work_output_path=None,
                                                keep_copies=False,
                                                overwrite=overwrite,
                                                logger=logger)

                # Clean up.
                #
                if not keep_intermediates:
                    if prediction_path != output_result_path:
                        dptimagefile.remove_image(image_path=prediction_path, ignore_errors=True, logger=logger)

                    if apply_zoom and zoomed_path != output_result_path:
                        dptimagefile.remove_image(image_path=zoomed_path, ignore_errors=True, logger=logger)

                    if apply_threshold and thresholded_path != output_result_path:
                        dptimagefile.remove_image(image_path=thresholded_path, ignore_errors=True, logger=logger)

                    if apply_diagonal_filter and filtered_path != output_result_path:
                        dptimagefile.remove_image(image_path=filtered_path, ignore_errors=True, logger=logger)

                    if apply_hole_filling and filled_path != output_result_path:
                        dptimagefile.remove_image(image_path=filled_path, ignore_errors=True, logger=logger)

                # Processing finished. Close the images.
                #
                input_image.close()
                if mask_image:
                    mask_image.close()

                # Copy the result to the target location.
                #
                if work_result_path:
                    if keep_copies:
                        dptimagefile.copy_image(source_path=work_result_path, target_path=target_result_path, overwrite=overwrite, logger=logger)
                    else:
                        dptimagefile.move_image(source_path=work_result_path, target_path=target_result_path, overwrite=overwrite, logger=logger)

                if work_interval_path:
                    if keep_copies:
                        dptimagefile.copy_image(source_path=work_interval_path, target_path=target_interval_path, overwrite=overwrite, logger=logger)
                    else:
                        dptimagefile.move_image(source_path=work_interval_path, target_path=target_interval_path, overwrite=overwrite, logger=logger)

                # Remove duplicated image files.
                #
                if not keep_copies:
                    if copy_image_path and copy_image_path != source_image_path:
                        dptimagefile.remove_image(image_path=copy_image_path, ignore_errors=True, logger=logger)

                    if copy_mask_path and copy_mask_path != source_mask_path:
                        dptimagefile.remove_image(image_path=copy_mask_path, ignore_errors=True, logger=logger)

                # Measure execution time.
                #
                segmentation_exec_time = time.time() - segmentation_start_time
                if logger:
                    logger.debug('Network application sequence done in {delta}'.format(delta=datetime.timedelta(seconds=segmentation_exec_time)))

            else:
                if logger:
                    logger.info('Skipping, target file already exits: {path}'.format(path=target_result_path))

        except Exception as exception:
            # Re-raise exception for the single file mode.
            #
            if single_mode:
                raise

            if logger:
                logger.error('Error: {exception}'.format(exception=exception))

            # Add case to the error collection.
            #
            failed_collection.append(target_result_path)

        else:
            # Save the result to the list of successful inferences.
            #
            successful_collection.append(target_result_path)

    # Log execution time.
    #
    if not single_mode:
        batch_execution_time = time.time() - batch_start_time
        if logger:
            logger.debug('Network application batch done in {delta}'.format(delta=datetime.timedelta(seconds=batch_execution_time)))

    # Return a list of successful and failed inferences.
    #
    return successful_collection, failed_collection

#----------------------------------------------------------------------------------------------------

def apply_network(input_path,
                  mask_path,
                  output_path,
                  interval_path,
                  model_path,
                  patch_size,
                  output_class,
                  number_of_classes,
                  normalizer,
                  normalizer_source_range,
                  normalizer_target_range,
                  soft_mode,
                  input_spacing,
                  output_spacing=None,
                  spacing_tolerance=0.25,
                  channels_first=False,
                  input_channels=(0, 1, 2),
                  confidence=0.5,
                  minimum_region_diagonal=0.0,
                  minimum_hole_diagonal=0.0,
                  full_connectivity=False,
                  quantize=True,
                  interpolation_order=0,
                  copy_input_path=None,
                  copy_mask_path=None,
                  work_output_path=None,
                  work_interval_path=None,
                  keep_intermediates=False,
                  keep_copies=False,
                  overwrite=True,
                  logger=None):
    """
    Apply network on image.

    Args:
        input_path (str): Path of the image to classify.
        mask_path (str): Path of the mask image to use.
        output_path (str): Path of the result image.
        interval_path (str, None): Quantization interval file path. If empty or none the interval will not be written out.
        model_path (str): Path of the model file to load.
        patch_size (int): Patch size to set for processing.
        output_class (int): Output class for soft classification. If negative all class predictions will be written out in separate channels.
        number_of_classes (int): Number of output classes. If negative will try to get number of channels of network.
        normalizer (str): Normalizer identifier.
        normalizer_source_range (tuple, None): Normalizer source range in (min, max) format.
        normalizer_target_range (tuple, None): Normalizer target range in (min, max) format.
        soft_mode (bool): Soft classification. If set too true the output will be the softmax.
        input_spacing (float): Pixel spacing level of WSI processed by the classifier.
        output_spacing (float, None): Required output pixel spacing: level 0 pixel spacing of the output image will be the given pixel spacing level of the input image.
        spacing_tolerance (float): Pixel spacing tolerance (percentage).
        channels_first (bool): If true, the network expect patches shaped BCHW, otherwise BHWC.
        input_channels (list): List of input channel indices.
        confidence (float): Network confidence for thresholding the predictions. Thresholding is omitted for threshold values smaller or equal to 0.0.
        minimum_region_diagonal (float): Region diagonal minimum (micrometer), for removing smaller than threshold sized regions from the output. Executed only if the diagonal is larger than 0.0.
        minimum_hole_diagonal (float): Hole diagonal minimum (micrometer), for filling smaller than threshold sized holes from the output. Executed only if the diagonal is larger than 0.0.
        full_connectivity (bool): Connectivity matrix for region and hole filtering. If true edge and point neighbors will be used otherwise edge neighbors only.
        quantize (bool): Quantization flag. If true the network output is quantized to [0, 255] uint8 interval.
        interpolation_order (int): Interpolation order from the [0, 5] interval.
        copy_input_path (str, None): Copy image path. If not None, the input image is copied to this path before processing.
        copy_mask_path (str, None): Copy mask path. If not None, the mask image is copied to this path before processing.
        work_output_path (str, None): Work output path. If not None, the result image is written here first before copying to the target path.
        work_interval_path (str, None): Work interval file path. If not None, the interval file is written here first before copying to the target path.
        keep_intermediates (bool): Keep intermediate files.
        keep_copies (bool): Keep the copied image files.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Raises:
        MissingModelFileError: Missing model file.
        UnknownImageNormalizationModeError: Unknown image normalization mode.
        InvalidPixelSpacingValueError: Both image level and pixel spacing are None.
        InvalidInterpolationOrderError: Interpolation order out of [0, 5] bounds.
        AsymmetricZoomFactorError: Asymmetric zoom values.
        InvalidZoomFactorError: Non-positive zoom factor.

        DigitalPathologyConfigError: Configuration errors.
        DigitalPathologyImageError: Image errors.
        DigitalPathologyModelError: Model errors.
        DigitalPathologyProcessingError: Processing errors.
    """

    # Call the batch mode function with a single job.
    #
    apply_network_batch(job_list=[(input_path, mask_path, copy_input_path, copy_mask_path,  work_output_path, work_interval_path, output_path, interval_path)],
                        model_path=model_path,
                        patch_size=patch_size,
                        output_class=output_class,
                        number_of_classes=number_of_classes,
                        normalizer=normalizer,
                        normalizer_source_range=normalizer_source_range,
                        normalizer_target_range=normalizer_target_range,
                        soft_mode=soft_mode,
                        input_spacing=input_spacing,
                        output_spacing=output_spacing,
                        spacing_tolerance=spacing_tolerance,
                        channels_first=channels_first,
                        input_channels=input_channels,
                        confidence=confidence,
                        minimum_region_diagonal=minimum_region_diagonal,
                        minimum_hole_diagonal=minimum_hole_diagonal,
                        full_connectivity=full_connectivity,
                        quantize=quantize,
                        interpolation_order=interpolation_order,
                        keep_intermediates=keep_intermediates,
                        keep_copies=keep_copies,
                        single_mode=True,
                        overwrite=overwrite,
                        logger=logger)
