"""
Functions in this module can calculate a classification preview by calculating an overlay image with the original and
the classification output mask. It can also save any level of a multi-resolution image as a plain image.
"""

from ..errors import processingerrors as dptprocessingerrors
from ..image import imagereader as dptimagereader
from ..utils import imagefile as dptimagefile

import numpy as np
import imageio
import os
import datetime
import time

#----------------------------------------------------------------------------------------------------

def calculate_preview(image,
                      mask,
                      preview_path,
                      level,
                      pixel_spacing,
                      spacing_tolerance,
                      alpha,
                      palette=None,
                      copy_image_path=None,
                      copy_mask_path=None,
                      work_preview_path=None,
                      keep_copies=False,
                      overwrite=True,
                      logger=None):
    """
    Create a preview image from the given image and mask on the given level by overlaying the original RGB image with a false colored mask label image (typically the output of the network inference).
    The palette of the coloring can be set. By default the mask value 0 is transparent, 1 is red, 2 is green, 3 is blue, 4 is yellow, etc. This function can be used to generate preview for the
    output of a network on a high image level to quickly review the results.

    Args:
        image (dptimagereader.ImageReader, str): Input image object or path.
        mask (dptimagereader.ImageReader, str): Input mask object or path.
        preview_path (str): Path of the result preview image file.
        level (int, None): Processing level of the image. Either this or the pixel_spacing parameter is required.
        pixel_spacing (float, None): Pixel spacing of the image to process. (micrometer). Either this or the level parameter is required.
        spacing_tolerance (float, None): Pixel spacing tolerance (percentage).
        alpha (float): Overlay alpha value for preview generation.
        palette (list, None): List of palette colors. Each color must be a tuple with 3 elements from the [0, 255] range.
        copy_image_path (str, None): Copy image path. If not None, the source image is copied to this path before processing. Not used if the input is an image object.
        copy_mask_path (str, None):  Copy mask path. If not None, the source mask is copied to this path before processing. Not used if the mask is an image object.
        work_preview_path (str, None): Work preview path. If not None, the result preview image created here and copied to the target preview path after processing.
        keep_copies (bool): Keep the copied image files.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Raises:
        InvalidPixelSpacingValueError: Both image level and pixel spacing are None.
        InvalidColorPaletteError: Invalid palette for mask image coloring.
        NoMatchingLevelInMaskImageError: No matching level found in the mask image for the RGB image.

        DigitalPathologyConfigError: Configuration errors.
        DigitalPathologyImageError: Image errors.
    """

    # Start time measurement.
    #
    start_time = time.time()

    # Check level and pixel spacing. One must be valid.
    #
    if level is None and pixel_spacing is None:
        raise dptprocessingerrors.InvalidPixelSpacingValueError()

    # Check the palette.
    #
    if palette:
        if len(palette) < 1 or any(len(color) != 3 or any(color_item < 0 or 255 < color_item for color_item in color) for color in palette):
            raise dptprocessingerrors.InvalidColorPaletteError(palette)

    # Construct source image paths and direct output paths.
    #
    if isinstance(image, dptimagereader.ImageReader):
        input_image_path = image.path
    else:
        input_image_path = copy_image_path if copy_image_path else image

    if isinstance(mask, dptimagereader.ImageReader):
        input_mask_path = mask.path
    else:
        input_mask_path = copy_mask_path if copy_mask_path else mask

    output_preview_path = work_preview_path if work_preview_path else preview_path

    # Log parameters.
    #
    if logger:
        logger.info('Calculating preview...')
        logger.info('Image: {path}'.format(path=image.path if isinstance(image, dptimagereader.ImageReader) else image))
        logger.info('Mask: {path}'.format(path=mask.path if isinstance(mask, dptimagereader.ImageReader) else mask))
        logger.info('Preview: {path}'.format(path=preview_path))

        if level is None:
            logger.info('Processing spacing: {spacing} um'.format(spacing=pixel_spacing))
        else:
            logger.info('Processing level: {level}'.format(level=level))

    # Check if the target image already exits.
    #
    if not os.path.isfile(preview_path) or overwrite:
        # Copy the source image and mask to the intermediate locations for processing.
        #
        if not isinstance(image, dptimagereader.ImageReader) and copy_image_path:
            dptimagefile.copy_image(source_path=image, target_path=copy_image_path, overwrite=overwrite, logger=logger)

        if not isinstance(mask, dptimagereader.ImageReader) and copy_mask_path:
            dptimagefile.copy_image(source_path=mask, target_path=copy_mask_path, overwrite=overwrite, logger=logger)

        # Create work directory structure if necessary.
        #
        if work_preview_path:
            os.makedirs(os.path.dirname(work_preview_path), exist_ok=True)

        # Open input image and mask image.
        #
        input_image = image if isinstance(image, dptimagereader.ImageReader) else dptimagereader.ImageReader(image_path=input_image_path, spacing_tolerance=spacing_tolerance, input_channels=None)
        input_mask = mask if isinstance(mask, dptimagereader.ImageReader) else dptimagereader.ImageReader(image_path=input_mask_path, spacing_tolerance=spacing_tolerance, input_channels=None)

        # Calculate source level and target spacing and add missing spacing information.
        #
        processing_level = input_image.level(spacing=pixel_spacing) if level is None else level

        if logger:
            if level is None:
                logger.debug('Identified level: {level}'.format(level=processing_level))

        if any(level_spacing is None for level_spacing in input_mask.spacings):
            input_mask.correct(spacing=1.0, level=0)

        # Find matching level for the mask with the image.
        #
        mask_level = None
        for level_index in range(len(input_mask.shapes)):
            if input_image.shapes[processing_level] == input_mask.shapes[level_index]:
                mask_level = level_index
                break

        if mask_level is None:
            raise dptprocessingerrors.NoMatchingLevelInMaskImageError(input_mask.path, input_image.path, processing_level)

        if logger:
            logger.info('Calculating preview: {path}'.format(path=output_preview_path))

        # Construct default palette for preview: red, green, blue, orange, magenta, purple, yellow, pink, teal
        #
        if palette:
            preview_palette = [[palette_item[0] / 255.0, palette_item[1] / 255.0, palette_item[2] / 255.0] for palette_item in palette]
        else:
            preview_palette = [[0.0, 0.0, 0.0],  # black (background)
                               [1.0, 0.0, 0.0],  # red
                               [0.0, 1.0, 0.0],  # green
                               [0.0, 0.0, 1.0],  # blue
                               [1.0, 0.5, 0.0],  # orange
                               [1.0, 0.0, 0.5],  # magenta
                               [0.5, 0.0, 1.0],  # purple
                               [1.0, 1.0, 0.0],  # yellow
                               [1.0, 0.5, 1.0],  # pink
                               [0.0, 1.0, 1.0],  # teal
                               [0.5, 0.0, 0.0]]  # brown

        # Convert the palette to numpy array for coloring.
        #
        palette_array = np.asarray(preview_palette, dtype=np.float32)

        # Load the images to numpy arrays.
        #
        image_array = input_image.content(spacing=input_image.spacings[processing_level])
        image_array = image_array.transpose((1, 2, 0))

        mask_array = input_mask.content(spacing=input_mask.spacings[mask_level])
        mask_array = mask_array.transpose((1, 2, 0)).squeeze()

        # Create triple channel alpha array.
        #
        alpha_array = np.full(shape=mask_array.shape, fill_value=alpha, dtype=np.float)
        alpha_array[mask_array == 0] = 0.0
        alpha_array = alpha_array[:, :, None]
        alpha_array = np.repeat(a=alpha_array, repeats=3, axis=2)

        # Compose overlay and save image.
        #
        image_array_float = image_array / 255.0
        mask_rgb_array_float = palette_array[mask_array]
        image_overlaid_array = image_array_float * (1.0 - alpha_array) + mask_rgb_array_float * alpha_array

        image_overlaid_array *= 255.0
        image_overlaid_array = image_overlaid_array.astype(dtype=np.uint8)

        imageio.imwrite(uri=output_preview_path, im=image_overlaid_array)

        # Processing finished. Close the images.
        #
        if not isinstance(mask, dptimagereader.ImageReader):
            input_mask.close()

        if not isinstance(image, dptimagereader.ImageReader):
            input_image.close()

        # Copy the result to the target location.
        #
        if work_preview_path:
            if keep_copies:
                dptimagefile.copy_image(source_path=work_preview_path, target_path=preview_path, overwrite=overwrite, logger=logger)
            else:
                dptimagefile.move_image(source_path=work_preview_path, target_path=preview_path, overwrite=overwrite, logger=logger)

        # Remove duplicated image files.
        #
        if not keep_copies:
            if not isinstance(image, dptimagereader.ImageReader) and copy_image_path and copy_image_path != image:
                dptimagefile.remove_image(image_path=copy_image_path, ignore_errors=True, logger=logger)

            if not isinstance(mask, dptimagereader.ImageReader) and copy_mask_path and copy_mask_path != mask:
                dptimagefile.remove_image(image_path=copy_mask_path, ignore_errors=True, logger=logger)

        # Log execution time.
        #
        execution_time = time.time() - start_time
        if logger:
            logger.debug('Done in {delta}'.format(delta=datetime.timedelta(seconds=execution_time)))

    else:
        if logger:
            logger.info('Skipping, target file already exits: {path}'.format(path=preview_path))

#----------------------------------------------------------------------------------------------------

def calculate_preview_batch(job_list, level, pixel_spacing, spacing_tolerance, alpha, palette=None, keep_copies=False, overwrite=True, logger=None):
    """
    Create image and mask overlay previews.

    Args:
        job_list (list): List of job (source image path, source mask path, copy image path, copy mask path, work preview path, target preview path) sextets. Source image, source mask and target
            preview paths are mandatory, the rest is optional. If copy paths are set, the source images are copied to the copy paths before processing. If work path is set, the result preview
            is written to the work path first then copied to the target preview path after processing.
        level (int, None): Processing level of the image. Either this or the pixel_spacing parameter is required.
        pixel_spacing (float, None): Pixel spacing of the image to process. (micrometer). Either this or the level parameter is required.
        spacing_tolerance (float, None): Pixel spacing tolerance (percentage).
        alpha (float): Overlay alpha value for preview generation.
        palette (list, None): List of palette colors. Each color must be a tuple with 3 elements from the [0, 255] range.
        keep_copies (bool): Keep the copied image files.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Returns:
        (list, list): List of successfully processed items and list of failed items.
    """

    # Start time measurement.
    #
    start_time = time.time()

    # Log batch processing.
    #
    if logger:
        logger.info('Calculating previews in batch mode...')
        logger.info('Job count: {count}'.format(count=len(job_list)))

    # Go through the job list.
    #
    failed_collection = []
    successful_collection = []
    for job_index in range(len(job_list)):
        # Get one job item.
        #
        source_image_path, source_mask_path, copy_image_path, copy_mask_path, work_preview_path, target_preview_path = job_list[job_index]

        try:
            # Print data paths if the mode is not a single file mode.
            #
            if logger:
                logger.info('Processing [{index}/{count}]: {path}'.format(index=job_index+1, count=len(job_list), path=target_preview_path))

            # Calculate the preview.
            #
            calculate_preview(image=source_image_path,
                              mask=source_mask_path,
                              preview_path=target_preview_path,
                              level=level,
                              pixel_spacing=pixel_spacing,
                              spacing_tolerance=spacing_tolerance,
                              alpha=alpha,
                              palette=palette,
                              copy_image_path=copy_image_path,
                              copy_mask_path=copy_mask_path,
                              work_preview_path=work_preview_path,
                              keep_copies=keep_copies,
                              overwrite=overwrite,
                              logger=logger)

        except Exception as exception:
            if logger:
                logger.error('Error: {exception}'.format(exception=exception))

            # Add case to the error collection.
            #
            failed_collection.append(target_preview_path)

        else:
            # Save the image to the list of successful calculations.
            #
            successful_collection.append(target_preview_path)

    # Log execution time.
    #
    execution_time = time.time() - start_time
    if logger:
        logger.debug('Preview calculation batch done in {delta}'.format(delta=datetime.timedelta(seconds=execution_time)))

    # Return a list of successful calculations.
    #
    return successful_collection, failed_collection

#----------------------------------------------------------------------------------------------------

def save_mrimage_as_image(image, output_path, level, pixel_spacing, spacing_tolerance, multiplier=1, overwrite=True, logger=None):
    """
    Save image at the required level.

    Args:
        image (dptimagereader.ImageReader, str): Input image object or path.
        output_path (str): Output image path.
        level (int, None): Processing level. Either this or the pixel_spacing parameter is required.
        pixel_spacing (float, None): Processing pixel spacing (micrometer). Either this or the level parameter is required.
        spacing_tolerance (float, None): Pixel spacing tolerance (percentage).
        multiplier (int): Multiplier to apply on the image. Might be useful in case of label images.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Raises:
        InvalidPixelSpacingValueError: Both image level and pixel spacing are None.

        DigitalPathologyConfigError: Configuration errors.
        DigitalPathologyImageError: Image errors.
    """

    # Start time measurement.
    #
    start_time = time.time()

    # Check level and pixel spacing. One must be valid.
    #
    if level is None and pixel_spacing is None:
        raise dptprocessingerrors.InvalidPixelSpacingValueError()

    # Log paths.
    #
    if logger:
        logger.info('Saving multiresolution image as plain image...')
        logger.info('Multiresolution image: {path}'.format(path=image.path if isinstance(image, dptimagereader.ImageReader) else image))
        logger.info('Plain image: {path}'.format(path=output_path))
        logger.info('Multiplier: {multiplier}'.format(multiplier=multiplier))

        if level is None:
            logger.info('Processing spacing: {spacing} um'.format(spacing=pixel_spacing))
        else:
            logger.info('Processing level: {level}'.format(level=level))

    # Check if the target image already exits.
    #
    if not os.path.isfile(output_path) or overwrite:
        # Open images.
        #
        input_image = image if isinstance(image, dptimagereader.ImageReader) else dptimagereader.ImageReader(image_path=image, spacing_tolerance=spacing_tolerance, input_channels=None)

        # Calculate processing level and add missing spacing information.
        #
        processing_level = input_image.level(spacing=pixel_spacing) if level is None else level

        if logger:
            if level is None:
                logger.debug('Identified level: {level}'.format(level=processing_level))

        if any(level_spacing is None for level_spacing in input_image.spacings):
            input_image.correct(spacing=1.0, level=0)

        # Collect size.
        #
        image_array = input_image.content(spacing=input_image.spacings[processing_level])
        image_array = image_array.transpose((1, 2, 0)).squeeze()

        if multiplier != 1:
            image_array *= multiplier

        # Configure the image writer.
        #
        imageio.imwrite(uri=output_path, im=image_array)

        # Close the input image.
        #
        if not isinstance(image, dptimagereader.ImageReader):
            input_image.close()

        # Log execution time.
        #
        execution_time = time.time() - start_time
        if logger:
            logger.debug('Done in {delta}'.format(delta=datetime.timedelta(seconds=execution_time)))

    else:
        if logger:
            logger.info('Skipping, target file already exits: {path}'.format(path=output_path))

#----------------------------------------------------------------------------------------------------

def save_mrimage_as_image_batch(job_list, level, pixel_spacing, spacing_tolerance, multiplier=1, overwrite=True, logger=None):
    """
    Save image at the required level.

    Args:
        job_list (list): List of job (input path, output path) pairs.
        level (int, None): Processing level. Either this or the pixel_spacing parameter is required.
        pixel_spacing (float, None): Processing pixel spacing (micrometer). Either this or the level parameter is required.
        spacing_tolerance (float, None): Pixel spacing tolerance (percentage).
        multiplier (int): Multiplier to apply on the image. Might be useful in case of label images.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Returns:
        (list, list): List of successfully processed items and list of failed items.
    """

    # Start time measurement.
    #
    start_time = time.time()

    # Log batch processing.
    #
    if logger:
        logger.info('Saving multiresolution images as plain images in batch mode...')
        logger.info('Job count: {count}'.format(count=len(job_list)))

    # Go through the job list.
    #
    failed_collection = []
    successful_collection = []
    for job_index in range(len(job_list)):
        # Get one job item.
        #
        mrimage_path, output_path = job_list[job_index]

        try:
            # Print data paths if the mode is not a single file mode.
            #
            if logger:
                logger.info('Processing [{index}/{count}]: {path}'.format(index=job_index+1, count=len(job_list), path=output_path))

            # Execute the conversion.
            #
            save_mrimage_as_image(image=mrimage_path,
                                  output_path=output_path,
                                  level=level,
                                  pixel_spacing=pixel_spacing,
                                  spacing_tolerance=spacing_tolerance,
                                  multiplier=multiplier,
                                  overwrite=overwrite,
                                  logger=logger)

        except Exception as exception:
            if logger:
                logger.error('Error: {exception}'.format(exception=exception))

            # Add case to the error collection.
            #
            failed_collection.append(output_path)

        else:
            # Save the result to the list of successful conversions.
            #
            successful_collection.append(output_path)

    # Log execution time.
    #
    execution_time = time.time() - start_time
    if logger:
        logger.debug('Image saving batch done in {delta}'.format(delta=datetime.timedelta(seconds=execution_time)))

    # Return a list of successful conversions.
    #
    return successful_collection, failed_collection
