"""
This file contains the reproduction of the Foreground Extraction from Structure Information (FESI) segmentation method from the paper
D. Bug, F. Feuerhake, and D. Merhof. 'Foreground extraction for histopathological whole-slide imaging.', Bildverarbeitung für die Medizin 2015,
pp. 419–424, 2015.
"""

import digitalpathology.image.imagereader as dptimagereader
import digitalpathology.image.imagewriter as dptimagewriter
import digitalpathology.processing.regions as dptregions
import digitalpathology.utils.imagefile as dptimagefile

import cv2
import numpy as np
import scipy.ndimage
import datetime
import time
import os
import yaml
import math

#----------------------------------------------------------------------------------------------------

def _close_points(seeds, start, distance_threshold):
    """
    Check if the seed points are close enough to the start.

    Args:
        seeds (list): Seed points.
        start (tuple): Start point.
        distance_threshold (float): Distance threshold.

    Returns:
        bool: True if the seed points close enough, false otherwise.
    """

    for seed in seeds:
        if (np.power(seed[0]-start[0], 2) + np.power(seed[1]-start[1], 2) < np.power(3*seed[2], 2)) and (seed[2] > distance_threshold):
            return True

    return False

#----------------------------------------------------------------------------------------------------

def _fesi_select_foreground(image_rgb_array,
                            gaussian_kernel_size=15,
                            gaussian_sigma=4.0,
                            median_kernel_size=45,
                            opening_kernel_size=7,
                            opening_iterations=5,
                            seed_distance_map_mask_size=3,
                            distance_map_mask_size=5,
                            distance_threshold=100.0):
    """
    Select foreground with the described FESI method from the image and returns the ndarray representing the result.

    The steps are:
        1. Load the image and apply a bit of preprocessing to remove the completely back areas that are background and should be white.
        2. Apply Laplacian filter on the preprocessed image.
        3. Apply Gaussian blur to the absolute value of the Laplacian image.
        4. Apply median blur and then morphological opening on the Gaussian blurred image.
        5. Apply distance transformation on the inverse mask and select maximal value as seed point.
        6. Flood fill the mask of step 4. from the previously identified seed point with 0 fillinig value to select the initial mask.
        7. Apply distance transformation on the initial mask.
        8. Loop over the maximal distance values in the distance image of step 7.
        8.1     Select the maximal distance point as seed point.
        8.2     Clear the region marked by the seed point from the mask and the distance values.
        8.3     If current maximal distance is large enough (=the region is large enough) or it is close enough to any previous seed point
                    mark the region in the final mask as foreground.
        9. Finalize the maks by removing all non-marked regions.

    Args:
        image_rgb_array (np.ndarray): Input RGB image numpy array to process.
        gaussian_kernel_size (int): Kerenel size for Gaussian blurring.
        gaussian_sigma (float): Sigma for Gaussian blurring.
        median_kernel_size (int): Median filter kernel size.
        opening_kernel_size (int): Morphological opening kernel size.
        opening_iterations (int): Morphological opening iteration count.
        seed_distance_map_mask_size (int): Mask size for distance map calculation in seed point search.
        distance_map_mask_size (int): Mask size for distance map calculation.
        distance_threshold (float): Distance threshold to select the large enough regions.

    Returns:
        np.ndarray: Result mask.
    """

    # Build tuples from sizes that is required by OpenCV.
    #
    gaussian_kernel_shape = (gaussian_kernel_size, gaussian_kernel_size)
    opening_kernel_shape = (opening_kernel_size, opening_kernel_size)

    # Convert RGB image to gray-scale: Y = R * 0.299 + G * 0.587 + B * 0.114
    #
    image_array = cv2.cvtColor(image_rgb_array, code=cv2.COLOR_RGB2GRAY)

    # Get the initial foreground mask.
    #
    laplace = np.abs(cv2.Laplacian(image_array, ddepth=cv2.CV_32F))
    blurred = cv2.GaussianBlur(laplace, ksize=gaussian_kernel_shape, sigmaX=gaussian_sigma)
    blurred_mean = blurred.mean()
    _, mask = cv2.threshold(blurred, thresh=blurred_mean, maxval=150.0, type=cv2.THRESH_BINARY)
    mask = mask.astype(np.uint8)

    # Refine foreground mask with with morphological opening.
    #
    mask = cv2.medianBlur(mask, median_kernel_size) + 100
    mask = cv2.morphologyEx(mask, op=cv2.MORPH_OPEN, kernel=cv2.getStructuringElement(cv2.MORPH_ELLIPSE, opening_kernel_shape), iterations=opening_iterations)

    # Select a point from background with distance transformation for seed point for flood fill.
    #
    _, inverse_mask = cv2.threshold(blurred, thresh=blurred_mean, maxval=255.0, type=cv2.THRESH_BINARY_INV)
    inverse_mask = inverse_mask.astype(np.uint8)
    distance_map = cv2.distanceTransform(inverse_mask, distanceType=cv2.cv2.DIST_C, maskSize=seed_distance_map_mask_size)

    flood_seed = np.unravel_index(distance_map.argmax(), distance_map.shape)
    flood_seed = (flood_seed[1], flood_seed[0])

    # Apply flood fill on the mask from the selected background point for refined foreground mask.
    #
    floodfill_mask = np.zeros((mask.shape[0] + 2, mask.shape[1] + 2), dtype=np.uint8)
    cv2.floodFill(mask, floodfill_mask, seedPoint=flood_seed, newVal=0)
    mask[mask > 1] = 255

    # Go through the distinctive regions and select the large enough or close enough ones.
    #
    final_mask = mask.copy()

    distance = cv2.distanceTransform(mask, distanceType=cv2.cv2.DIST_C, maskSize=distance_map_mask_size)
    dist_max = distance.max()

    seeds = []
    while 0 < dist_max:
        start = np.unravel_index(distance.argmax(), distance.shape)
        start = (start[1], start[0])
        seeds.append((start[0], start[1], dist_max))

        floodfill_mask.fill(0)
        cv2.floodFill(mask, floodfill_mask, seedPoint=start, newVal=0)
        distance[mask == 0] = 0
        dist_max = distance.max()

        if (dist_max > distance_threshold) or _close_points(seeds, start, distance_threshold):
            floodfill_mask.fill(0)
            cv2.floodFill(final_mask, floodfill_mask, seedPoint=start, newVal=200)

    # Finalize the mask: clear everything that is not marked as foreground so far.
    #
    return np.equal(final_mask, 200)

#----------------------------------------------------------------------------------------------------

def _otsu_select_foreground(image_rgb_array):
    """
    Select foreground with Otsu's threshold.

    Args:
        image_rgb_array (np.ndarray): Input RGB image numpy array to process.

    Returns:
        np.ndarray: Result mask.
    """

    # Convert RGB image to gray-scale: Y = R * 0.299 + G * 0.587 + B * 0.114
    #
    image_array = cv2.cvtColor(image_rgb_array, code=cv2.COLOR_RGB2GRAY)

    threshold, result_array = cv2.threshold(src=image_array, thresh=0, maxval=255, type=cv2.THRESH_OTSU)

    # Return the thresholded image.
    #
    return np.equal(result_array, 0)

#----------------------------------------------------------------------------------------------------

def _threshold_select_foreground(image_rgb_array, threshold):
    """
    Select foreground with threshold.

    Args:
        image_rgb_array (np.ndarray): Input RGB image numpy array to process.
        threshold (int): Threshold value.

    Returns:
        np.ndarray: Result mask.
    """

    # Convert RGB image to gray-scale: Y = R * 0.299 + G * 0.587 + B * 0.114
    #
    image_array = cv2.cvtColor(image_rgb_array, code=cv2.COLOR_RGB2GRAY)

    # Apply threshold.
    #
    result_array = image_array < threshold

    # Refine the result with morphological closing and hole filling.
    #
    result_array = scipy.ndimage.morphology.binary_closing(input=result_array, structure=None, iterations=1)
    result_array = scipy.ndimage.morphology.binary_fill_holes(input=result_array)

    # Return the thresholded image.
    #
    return result_array

#----------------------------------------------------------------------------------------------------

def calculate_tissue_mask(image,
                          result_path,
                          pixel_spacing,
                          spacing_tolerance,
                          method,
                          gray_threshold,
                          minimum_region_diagonal,
                          minimum_hole_diagonal,
                          copy_image_path=None,
                          work_result_path=None,
                          keep_copies=False,
                          overwrite=True,
                          logger=None):
    """
    Calculate the tissue mask with Otsu's threshold or the FESI algorithm or smimple thresholding.

    Args:
        image (dptimagereader.ImageReader, str): Input image object or path.
        result_path (str): Result mask image path.
        pixel_spacing (float, None): Processing pixel spacing (micrometer).
        spacing_tolerance (float, None): Pixel spacing tolerance (percentage).
        method (str): Method identifier. Accepted values are: 'otsu', 'fesi' and 'threshold'.
        gray_threshold (int, None): Graysclae threshold for 'threshold' method.
        minimum_region_diagonal (float): Region diagonal minimum (micrometer), for removing smaller than threshold sized regions from the output. Executed only if the diagonal is larger than 0.0.
        minimum_hole_diagonal (float): Hole diagonal minimum (micrometer), for filling smaller than threshold sized holes from the output. Executed only if the diagonal is larger than 0.0.
        copy_image_path (str, None): Copy image path. If not None, the source input image is copied to this path before processing. Not used if the input image is an image object.
        work_result_path (str, None): Work result path. If not None, the result output image created here and copied to the target result path after processing.
        keep_copies (bool): Keep the copied annotation files.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.
    """

    # Start time measurement.
    #
    start_time = time.time()

    # Check processing method.
    #
    if method not in ['otsu', 'fesi', 'threshold']:
        raise ValueError('Unknown processing method: {method}'.format(method=method))

    if isinstance(image, dptimagereader.ImageReader):
        input_image_path = image.path
    else:
        input_image_path = copy_image_path if copy_image_path else image

    output_result_path = work_result_path if work_result_path else result_path

    # Log parameters.
    #
    if logger:
        logger.info('Calculating tissue mask...')
        logger.info('Input image: {path}'.format(path=image.path if isinstance(image, dptimagereader.ImageReader) else image))
        logger.info('Result image: {path}'.format(path=result_path))
        logger.info('Algorithm: {name}'.format(name=method))
        logger.info('Processing spacing: {spacing} um'.format(spacing=pixel_spacing))
        logger.info('Diagonal threshold: {threshold} um'.format(threshold=minimum_region_diagonal))

    # Check if the target image already exits.
    #
    if not os.path.isfile(result_path) or overwrite:
        # Copy the source input image to the intermediate location for processing.
        #
        if not isinstance(image, dptimagereader.ImageReader) and copy_image_path:
            dptimagefile.copy_image(source_path=image, target_path=copy_image_path, overwrite=overwrite, logger=logger)

        # Create work directory structure if necessary.
        #
        if work_result_path:
            os.makedirs(os.path.dirname(work_result_path), exist_ok=True)

        # Open images.
        #
        input_image = image if isinstance(image, dptimagereader.ImageReader) else dptimagereader.ImageReader(image_path=input_image_path, spacing_tolerance=spacing_tolerance, input_channels=None)

        # Calculate source level and target spacing and add missing spacing information.
        #
        processing_level = input_image.level(spacing=pixel_spacing)

        if logger:
            logger.debug('Identified level: {level}'.format(level=processing_level))

        # Load the image into memory.
        #
        if logger:
            logger.debug('Processing: {path}'.format(path=input_image.path))

        image_array = input_image.content(spacing=pixel_spacing).transpose((1, 2, 0))

        if method == 'fesi':
            fesi_level = input_image.level(spacing=8.0)
            fesi_distance = input_image.spacings[fesi_level] / input_image.spacings[processing_level] * 100.0

            if logger:
                logger.debug('FESI distance: {distance} pixels'.format(distance=fesi_distance))

            mask_array = _fesi_select_foreground(image_rgb_array=image_array, distance_threshold=fesi_distance)
        elif method == 'otsu':
            mask_array = _otsu_select_foreground(image_rgb_array=image_array)
        elif method == 'threshold':
            mask_array = _threshold_select_foreground(image_rgb_array=image_array, threshold=gray_threshold)
        else:
            raise ValueError('Unknown processing method: {method}'.format(method=method))

        # Filter the smaller than threshold regions out from the results.
        #
        if 0.0 < minimum_region_diagonal:
            if logger:
                logger.debug('Filtering small regions...')

            minimum_region_diagonal_pixels = minimum_region_diagonal / input_image.spacings[processing_level]

            if logger:
                logger.debug('Diagonal low threshold: {threshold} pixels'.format(threshold=minimum_region_diagonal_pixels))

            mask_array, total_regions, filtered_regions = dptregions.filter_regions_array(input_array=mask_array,
                                                                                          diagonal_threshold=minimum_region_diagonal_pixels,
                                                                                          full_connectivity=True,
                                                                                          foreground_labels=None,
                                                                                          background_label=0)

            if logger:
                logger.debug('Filtered: {filtered} of {total} regions'.format(filtered=filtered_regions, total=total_regions))

        # Fill the smaller than threshold holes in the results.
        #
        if 0.0 < minimum_hole_diagonal:
            if logger:
                logger.debug('Filling holes...')

            minimum_hole_diagonal_pixels = minimum_hole_diagonal / input_image.spacings[processing_level]

            if logger:
                logger.debug('Diagonal low threshold: {threshold} pixels'.format(threshold=minimum_hole_diagonal_pixels))

            mask_array, total_holes, filled_holes = dptregions.fill_holes_array(input_array=mask_array,
                                                                                diagonal_threshold=minimum_hole_diagonal_pixels,
                                                                                full_connectivity=True,
                                                                                foreground_labels=None,
                                                                                fill_value=1)

            if logger:
                logger.debug('Filled: {filled} of {total} holes'.format(filled=filled_holes, total=total_holes))

        # Write out the result to TIF file.
        #
        if logger:
            logger.debug('Writing: {path}'.format(path=output_result_path))

        tile_size = 512
        image_writer = dptimagewriter.ImageWriter(image_path=output_result_path,
                                                  shape=input_image.shapes[processing_level],
                                                  spacing=input_image.spacings[processing_level],
                                                  dtype=np.uint8,
                                                  coding='monochrome',
                                                  compression=None,
                                                  interpolation=None,
                                                  tile_size=tile_size,
                                                  jpeg_quality=None,
                                                  empty_value=0,
                                                  skip_empty=False)

        mask_array = mask_array[None, ...]
        mask_array = mask_array.astype(dtype=np.uint8)
        image_writer.fill(content=mask_array)

        # Finalize the output image and close the input image.
        #
        image_writer.close()

        if not isinstance(image, dptimagereader.ImageReader):
            input_image.close()

        # Copy the result to the target location.
        #
        if work_result_path:
            if keep_copies:
                dptimagefile.copy_image(source_path=work_result_path, target_path=result_path, overwrite=overwrite, logger=logger)
            else:
                dptimagefile.move_image(source_path=work_result_path, target_path=result_path, overwrite=overwrite, logger=logger)

        # Remove duplicated image files.
        #
        if not keep_copies:
            if not isinstance(image, dptimagereader.ImageReader) and copy_image_path and copy_image_path != image:
                dptimagefile.remove_image(image_path=copy_image_path, ignore_errors=True, logger=logger)

        # Log execution time.
        #
        execution_time = time.time() - start_time
        if logger:
            logger.debug('Done in {delta}'.format(delta=datetime.timedelta(seconds=execution_time)))

    else:
        if logger:
            logger.info('Skipping, target file already exits: {path}'.format(path=result_path))

#----------------------------------------------------------------------------------------------------

def calculate_tissue_mask_batch(job_list, pixel_spacing, spacing_tolerance, method, gray_threshold, minimum_region_diagonal, minimum_hole_diagonal, keep_copies=False, overwrite=True, logger=None):
    """
    Calculate tissue mask with Otus's method or the FESI algorithm or simple thresholding in batch mode.

    Args:
        job_list (list): List of job (source input path, copy input path, work output path, target output path) quartets.
        pixel_spacing (float, None): Processing pixel spacing (micrometer).
        spacing_tolerance (float, None): Pixel spacing tolerance (percentage).
        method (str): Method identifier. Accepted values are: 'otsu', 'fesi', and 'threshold'.
        gray_threshold (int, None): Graysclae threshold for 'threshold' method.
        minimum_region_diagonal (float): Region diagonal minimum (micrometer), for removing smaller than threshold sized regions from the output. Executed only if the diagonal is larger than 0.0.
        minimum_hole_diagonal (float): Hole diagonal minimum (micrometer), for filling smaller than threshold sized holes from the output. Executed only if the diagonal is larger than 0.0.
        keep_copies (bool): Keep the copied annotation files.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Returns:
        (list, list): List of successfully processed items and list of failed items.
    """

    # Go through the job list.
    #
    failed_collection = []
    successful_collection = []
    for job_index in range(len(job_list)):
        # Get one job item.
        #
        source_input_path, copy_input_path, work_output_path, target_output_path = job_list[job_index]

        try:
            # Print data paths if the mode is not a single file mode.
            #
            if logger:
                logger.info('Processing [{index}/{count}]: {path}'.format(index=job_index+1, count=len(job_list), path=source_input_path))

            # Calculate tissue mask with FESI or Otsu's method.
            #
            calculate_tissue_mask(image=source_input_path,
                                  result_path=target_output_path,
                                  pixel_spacing=pixel_spacing,
                                  spacing_tolerance=spacing_tolerance,
                                  method=method,
                                  gray_threshold=gray_threshold,
                                  minimum_region_diagonal=minimum_region_diagonal,
                                  minimum_hole_diagonal=minimum_hole_diagonal,
                                  copy_image_path=copy_input_path,
                                  work_result_path=work_output_path,
                                  keep_copies=keep_copies,
                                  overwrite=overwrite,
                                  logger=logger)

        except Exception as exception:
            if logger:
                logger.error('Error: {exception}'.format(exception=exception))

            # Add case to the error collection.
            #
            failed_collection.append(target_output_path)

        else:
            # Save the result to the list of successful exports.
            #
            successful_collection.append(target_output_path)

    # Return a list of successful exports.
    #
    return successful_collection, failed_collection

#----------------------------------------------------------------------------------------------------

def list_missed_regions(comparison, comparison_labels, copy_comparison_path=None, keep_copies=False, overwrite=False, logger=None):
    """

    Args:
        comparison:
        comparison_labels:
        copy_comparison_path:
        keep_copies:
        overwrite:
        logger:

    Returns:
        list:
    """

    # Start time measurement.
    #
    start_time = time.time()

    if isinstance(comparison, dptimagereader.ImageReader):
        input_comparison_path = comparison.path
    else:
        input_comparison_path = copy_comparison_path if copy_comparison_path else comparison

    # Log parameters.
    #
    if logger:
        logger.info('Listing missed regions...')
        logger.info('Comparison image: {path}'.format(path=comparison.path if isinstance(comparison, dptimagereader.ImageReader) else comparison))
        logger.info('Missed labels: {labels}'.format(labels=comparison_labels))

    # Copy the comparison image to the intermediate location for processing.
    #
    if not isinstance(comparison, dptimagereader.ImageReader) and copy_comparison_path:
        dptimagefile.copy_image(source_path=comparison, target_path=copy_comparison_path, overwrite=overwrite, logger=logger)

    # Open images.
    #
    comparison_image = comparison if isinstance(comparison, dptimagereader.ImageReader) else dptimagereader.ImageReader(image_path=input_comparison_path, spacing_tolerance=0.25, input_channels=None)

    comparison_content = comparison_image.content(spacing=comparison_image.spacings[0])
    comparison_content = comparison_content.squeeze()
    comparison_binary = np.isin(element=comparison_content, test_elements=comparison_labels)

    connectivity_structure = np.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], dtype=np.bool)
    comparison_label_array, _ = scipy.ndimage.measurements.label(input=comparison_binary, structure=connectivity_structure)
    comparison_objects = scipy.ndimage.measurements.find_objects(input=comparison_label_array)

    diagonal_list = []
    for object_index in range(len(comparison_objects)):
        object_bounding_box = comparison_objects[object_index]

        object_height = object_bounding_box[0].stop - object_bounding_box[0].start
        object_width = object_bounding_box[1].stop - object_bounding_box[1].start
        object_diagonal = math.sqrt(object_height * object_height + object_width * object_width)
        object_diagonal *= comparison_image.spacings[0]

        diagonal_list.append(object_diagonal)

    if logger:
        logger.info('Missed diagonals: {diagonals}'.format(diagonals=diagonal_list))

    # Remove duplicated image files.
    #
    if not keep_copies:
        if not isinstance(comparison, dptimagereader.ImageReader) and copy_comparison_path and copy_comparison_path != comparison:
            dptimagefile.remove_image(image_path=copy_comparison_path, ignore_errors=True, logger=logger)

    # Log execution time.
    #
    execution_time = time.time() - start_time
    if logger:
        logger.debug('Done in {delta}'.format(delta=datetime.timedelta(seconds=execution_time)))

    return diagonal_list

#----------------------------------------------------------------------------------------------------

def list_missed_regions_batch(job_list, comparison_labels, output_paths, keep_copies=False, overwrite=True, logger=None):
    """


    Args:
        job_list (list): List of job (source comparison path, copy comparison path) pairs.
        comparison_labels (list): List of missing label values.
        output_paths (str): (Work, target) output YAML file paths.
        keep_copies (bool): Keep the copied annotation files.
        overwrite (bool): If true existing targets will be overwritten.
        logger (logging.Logger, None): Logger object.

    Returns:
        (list, list): List of successfully processed items and list of failed items.
    """

    # Go through the job list.
    #
    failed_collection = []
    successful_collection = []

    result_map = {}
    work_output_path, target_output_path = output_paths

    if overwrite or not os.path.isfile(target_output_path):
        for job_index in range(len(job_list)):
            # Get one job item.
            #
            source_comparison_path, copy_comparison_path = job_list[job_index]

            try:
                # Print data paths if the mode is not a single file mode.
                #
                if logger:
                    logger.info('Processing [{index}/{count}]: {path}'.format(index=job_index+1, count=len(job_list), path=source_comparison_path))

                # Calculate tissue mask with FESI or Otsu's method.
                #
                result_map[os.path.basename(source_comparison_path)] = list_missed_regions(comparison=source_comparison_path,
                                                                                           comparison_labels=comparison_labels,
                                                                                           copy_comparison_path=copy_comparison_path,
                                                                                           keep_copies=keep_copies,
                                                                                           overwrite=overwrite,
                                                                                           logger=logger)

            except Exception as exception:
                if logger:
                    logger.error('Error: {exception}'.format(exception=exception))

                # Add case to the error collection.
                #
                failed_collection.append(source_comparison_path)

            else:
                # Save the result to the list of successful exports.
                #
                successful_collection.append(source_comparison_path)

    # Save the output.
    #
    output_path = work_output_path if work_output_path else target_output_path

    with open(file=output_path, mode='w') as output_file:
        yaml.dump(data=result_map, stream=output_file, indent=4, default_flow_style=False)

    if output_path != target_output_path:
        dptimagefile.copy_image(source_path=output_path, target_path=target_output_path, overwrite=overwrite, logger=logger)

    # Remove the output to the target path.
    #
    if work_output_path:
        if keep_copies:
            dptimagefile.copy_image(source_path=work_output_path, target_path=target_output_path, overwrite=overwrite, logger=logger)
        else:
            dptimagefile.move_image(source_path=work_output_path, target_path=target_output_path, overwrite=overwrite, logger=logger)

    # Return a list of successful exports.
    #
    return successful_collection, failed_collection
