"""
This module can retrieve basic information from a Git repository.
"""

import git
import re

#----------------------------------------------------------------------------------------------------

def git_info(repo_dir_path):
    """
    Get basic information about a Git repository.

    Args:
        repo_dir_path (str): Directory path of a repository.

    Returns:
        dict, None: A dictionary of the retrieved information with 'name', 'url', 'revision' and 'branch' keys or None in case of error.
    """

    # Build a repository object and obtain basic information: the URL and the revision-branch pair.
    #
    try:
        repo = git.Repo(repo_dir_path)
    except (git.exc.InvalidGitRepositoryError, git.exc.NoSuchPathError):
        return None
    else:
        repo_url = repo.remotes.origin.url
        repo_name_rev = repo.head.object.name_rev

    # Find the name of the repository in the URL.
    #
    repo_name_match = re.match(pattern='.*/(?P<name>.*)(\.git)?$', string=repo_url)
    if repo_name_match:
        # Extract name from the match.
        #
        repo_name = repo_name_match.group('name')

        # Find the hed revision of the repository and the name of the active branch.
        #
        repo_revision_branch_match = re.match(pattern='^(?P<revision>[0-9a-f]*) (?P<branch>.*)$', string=repo_name_rev)
        repo_revision = repo_revision_branch_match.group('revision') if repo_revision_branch_match else None
        repo_branch = repo_revision_branch_match.group('branch') if repo_revision_branch_match else None

        # Construct the dictionary of the collected values.
        #
        return {'name': repo_name, 'url': repo_url, 'revision': repo_revision, 'branch': repo_branch}

    else:
        # Repository name matching failed.
        #
        return None
