"""
This file contains utility functions that update a saved network to the newest format.
"""

import digitalpathology.errors.modelerrors as dptmodelerrors

import h5py
import os

#----------------------------------------------------------------------------------------------------

def hf5_memory_file(name, mode):
    """
    Create a purely in-memory HF5 file that does not synchronize to any file on disk.

    Args:
        name (str): Name of the file. It still has to be set despite it is not synchronized to disk.
        mode (str): Read/write mode.

    Returns:
        h5py.File: The constructed HF5 file.
    """

    return h5py.File(name=name, mode=mode, driver='core', backing_store=False)

#----------------------------------------------------------------------------------------------------

def hf5_memory_file_image(hf5_file):
    """
    Get the byte content of the HF5 file.

    Args:
        hf5_file (h5py.File): HF5 file.

    Returns:
        bytes: The byte content of the HF5 file.
    """

    return h5py.h5f.FileID.get_file_image(hf5_file.id)

#----------------------------------------------------------------------------------------------------

def load_bytes_to_hf5_memory_file(byte_stream, name):
    """
    Loads a HF5 file from a byte stream.

    Args:
        byte_stream (bytes): Byte content of the HF5 file.
        name (str): Name of the file.

    Returns:
        h5py.File: The HF5 file loaded from the byte stream.
    """

    # Create a file access property ist for the HF5 file with in-memory representation and set the stored binary data to it.
    #
    file_access_property_list = h5py.h5p.create(h5py.h5p.FILE_ACCESS)
    file_access_property_list.set_fapl_core(backing_store=False)
    file_access_property_list.set_file_image(byte_stream)

    # Construct the H5 file with the settings. The filename is mandatory even though the content will not be read or written to the file at any point.
    #
    file_id = h5py.h5f.open(name=name.encode(), flags=h5py.h5f.ACC_RDONLY, fapl=file_access_property_list)
    hf5_file = h5py.File(name=file_id)

    # Return the constructed HF5 file.
    #
    return hf5_file

#----------------------------------------------------------------------------------------------------

def free_model_input_shape(model):
    """
    Free up the fixed input shape in a Keras model. 1. If the shape of the input layer is shaped (e.g. (128, 128, 3) instead of (None, None, 3)) then the function replaces the
    input layer and rebuilds the layer topology from the ground by copying the parameters of each further layer. 2. The function can also remove the last Reshape layer. It might
    be necessary if the 'predict' function returns the output of the Reshape layer.

    Args:
        model (keras.models.Model): Keras model to modify.

    Returns:
        keras.models.Model: Modified model.

    Raises:
        NotSupportedLayerError: Not supported layer encountered.
    """

    # Import Keras locally to avoid unnecessary GPU allocation.
    #
    import keras

    # If any of the changes has to be made the whole layer topology has to be rebuild.
    #
    old_input_shape = model.layers[0].input_shape

    if any(size is not None for size in old_input_shape[:-1]) or isinstance(model.layers[-1], keras.layers.Reshape):
        # Calculate the new input shape and construct a new input layer: the shape of the input layer also contains the batch size.
        #
        new_input_shape = (None,) * (len(old_input_shape) - 2) + (old_input_shape[-1],)
        new_input_layer = keras.layers.Input(shape=new_input_shape)
        stop_layer_index = len(model.layers) - 1 if isinstance(model.layers[-1], keras.layers.Reshape) else len(model.layers)

        # Rebuild the the layer topology.
        #
        last_new_layer = new_input_layer
        for index in range(1, stop_layer_index):
            # Deep copy layers.
            #
            layer_config = model.layers[index].get_config()

            # Layers with shape information are not compatible (e.g. Reshape layers).
            #
            if 'target_shape' in layer_config:
                raise dptmodelerrors.NotSupportedLayerError(model.layers[index])

            new_layer = keras.layers.deserialize(config={'class_name': model.layers[index].__class__.__name__, 'config': layer_config})
            last_new_layer = new_layer(last_new_layer)

        # Build new model.
        #
        new_model = keras.models.Model(inputs=new_input_layer, outputs=last_new_layer)

        # Sync parameters.
        #
        for index in range(1, len(new_model.layers)):
            new_model.layers[index].set_weights(weights=model.layers[index].get_weights())

        # Return the new model.
        #
        return new_model

    else:
        # No changes has to be made.
        #
        return model

#----------------------------------------------------------------------------------------------------

def free_network_input_shape(network):
    """
    Load the model from the network data and unfix the shape if the input layer and remove the last layer if it is a Reshape layer.

    Args:
        network (dict): Network data structure.

    Returns:
        dict: The updated network data structure.
    """

    # Import Keras locally to avoid unnecessary GPU allocation.
    #
    import keras

    # Build the model. Use unique file name to avoid conflicts.
    #
    model_hf5_name = '{module_name}.{function_name}.{item_name}.hf5'.format(module_name=__name__, function_name=free_network_input_shape.__name__, item_name='original_model')
    model_hf5 = load_bytes_to_hf5_memory_file(byte_stream=network['parameters']['hf5_image'], name=model_hf5_name)
    model = keras.models.load_model(filepath=model_hf5)
    model_hf5.close()

    # Free the input shape in the model.
    #
    new_model = free_model_input_shape(model=model)

    # The optimizer cannot be saved if the model topology has changed (e.g. a Reshape layer has been removed).
    #
    save_optimizer = len(model.layers) == len(new_model.layers)

    # Create a purely in-memory HF5 file for saving the model and save the new model to the memory file. Use unique file name to avoid conflicts.
    #
    new_model_hf5_name = '{module_name}.{function_name}.{item_name}.hf5'.format(module_name=__name__, function_name=free_network_input_shape.__name__, item_name='updated_model')
    new_model_hf5 = hf5_memory_file(name=new_model_hf5_name, mode='w')
    new_model.save(filepath=new_model_hf5, overwrite=True, include_optimizer=save_optimizer)

    # Get the file content into a byte array and return it.
    #
    network['parameters']['hf5_image'] = hf5_memory_file_image(hf5_file=new_model_hf5)
    new_model_hf5.close()

    return network

#----------------------------------------------------------------------------------------------------

def update_network_format(network):
    """
    Update a network with a possibly missing 'stamp' key and copy the HF5 file content into the network data structure as byte stream. For opening the HF5 file
    the path of the original network file must be injected in the data dictionary under a 'path' key.

    Args:
        network (dict): Network data structure.

    Returns:
        dict: Updated network data structure.
    """

    # Remove 'stamp' key if present.
    #
    if 'stamp' in network:
        del network['stamp']

    # Store the HF5 file content inside the model file as byte stream.
    #
    if type(network['parameters']) is dict and 'keras_model_file_name' in network['parameters']:
        # Construct the HF4 file path. Some old networks stored it as an absolute path. It is expected to be next to the network file.
        #
        hf5_path = '{base}{ext}'.format(base=os.path.splitext(network['path'])[0], ext='.hf5')

        # Load the file into memory and extract it as a byte stream.
        #
        model_hf5 = h5py.File(name=hf5_path, mode='r', driver='core', backing_store=False)
        model_hf5_bytes = h5py.h5f.FileID.get_file_image(model_hf5.id)

        # Remove the path to the HF5 file from the network parameters and insert it as byte stream.
        #
        del network['parameters']['keras_model_file_name']
        network['parameters']['hf5_image'] = model_hf5_bytes

    # Return the updated network.
    #
    return network

#----------------------------------------------------------------------------------------------------

def export_model_to_hf5(network, hf5_path):
    """
    Export the Keras model data to a separate HF5 file.

    Args:
        network (dict): Network data structure.
        hf5_path (str): Target HF5 file path.

    Returns:
        bool: True if the HF5 file is successfully exported, False otherwise.
    """

    # The function only works with the newest format of Keras networks.
    #
    if type(network['parameters']) is dict and 'keras_model_file_name' in network['parameters']:
        # Write out the binary content of the HF5 file to the target file.
        #
        with open(file=hf5_path, mode='wb') as hf5_file:
            hf5_file.write(network['parameters']['hf5_image'])

            # File successfully saved.
            #
            return True

    # The data is missing or the network is not a Keras network.
    #
    return False
