"""
This module can add internal and external margins with new labels to an existing region in a multi-resolution mask image file.
"""

import digitalpathology.processing.regions as dptregions
import digitalpathology.utils.foldercontent as dptfoldercontent
import digitalpathology.utils.loggers as dptloggers

import argparse
import os
import sys

#----------------------------------------------------------------------------------------------------

def compile_job_item_with_replacements(input_path, output_path, copy_directory, work_directory):
    """
    Finalize job item with copy and work directories.

    Args:
        input_path (str): Input path filter expression.
        output_path (str): Output path filter expression with possible {image} replacement key.
        copy_directory (str, None): Copy directory path where the image and mask images are copied before processing.
        work_directory (str, None): Work directory path where the preview image is written first before copying to the target path.

    Returns:
        tuple: A job tuple expanded with copy and work directories.
    """

    # Use the intermediate directory paths.
    #
    source_input_path = input_path

    if copy_directory:
        copy_input_path = os.path.join(copy_directory, os.path.basename(input_path))
    else:
        copy_input_path = None

    if work_directory:
        work_output_path = os.path.join(work_directory, os.path.basename(output_path))
    else:
        work_output_path = None

    target_output_path = output_path

    # Return assembled tuple.
    #
    return source_input_path, copy_input_path, work_output_path, target_output_path

#----------------------------------------------------------------------------------------------------

def assemble_jobs(input_path, output_path, copy_directory, work_directory):
    """
    Assemble (image path, annotation path, output path) job triplets.

    Args:
        input_path (str): Input path filter expression.
        output_path (str): Output path expression with possible {image} replacement key.
        copy_directory (str, None): Copy directory path where the input image is copied before processing.
        work_directory (str, None): Work directory path where the output image is written first before copying to the target path.

    Returns:
        list: List of job tuples.
    """

    # Find out operation mode. If all paths are file paths the mode is 'file'.
    #
    file_mode = os.path.isfile(input_path)
    print('Mode: {mode}'.format(mode=('file' if file_mode else 'folder')))

    result_job_list = []
    if file_mode:
        # Return a single triplet if the paths were existing files.
        #
        input_base = os.path.splitext(os.path.basename(input_path))[0]
        target_output_path = output_path.format(input=input_base)

        # Add intermediate paths.
        #
        job_item = compile_job_item_with_replacements(input_path=input_path,
                                                      output_path=target_output_path,
                                                      copy_directory=copy_directory,
                                                      work_directory=work_directory)

        # Add the assembled job item to the list.
        #
        result_job_list.append(job_item)
    else:
        # Collect all source images and annotations and match their file name for processing.
        #
        input_file_path_list = dptfoldercontent.folder_content(folder_path=input_path, recursive=False)

        # Assemble list.
        #
        for input_path_item in input_file_path_list:
            # Calculate output path.
            #
            input_base = os.path.splitext(os.path.basename(input_path_item))[0]
            target_output_path = output_path.format(input=input_base)

            # Add intermediate paths.
            #
            job_item = compile_job_item_with_replacements(input_path=input_path_item,
                                                          output_path=target_output_path,
                                                          copy_directory=copy_directory,
                                                          work_directory=work_directory)

            # Add the assembled job item to the list.
            #
            result_job_list.append(job_item)

        # Print match count for checking.
        #
        print('Matching mask count: {match_count}'.format(match_count=len(result_job_list)))

    # Return the result list.
    #
    return result_job_list

#----------------------------------------------------------------------------------------------------

def collect_arguments():
    """
    Collect command line arguments.

    Returns:
        (str, str, int, int, int, float, float, float, str, str, bool, bool): The parsed command line arguments: input mask path, output mask path, source label value, internal margin label
            value, external margin label value, margin size, input pixel spacing, pixel spacing tolerance, target directory path for data copy, work directory path, keep copied files flag,
            and the overwrite flag.
    """

    # Configure argument parser.
    #
    argument_parser = argparse.ArgumentParser(description='Add inner and outer margins to regions.',
                                              epilog='It can either work with file paths and file filter expressions. If filter expressions are set the matching is based on file names. '
                                                     'Use {input} as replacement for input file base name in output file specification.')

    argument_parser.add_argument('-i',  '--input',          required=True,  type=str,                 help='input mask or directory')
    argument_parser.add_argument('-o',  '--output',         required=True,  type=str,                 help='output mask or directory')
    argument_parser.add_argument('-l',  '--label',          required=True,  type=int,                 help='label to process')
    argument_parser.add_argument('-n',  '--internal',       required=True,  type=int,                 help='label value of the internal margin')
    argument_parser.add_argument('-e',  '--external',       required=True,  type=int,                 help='label value of the external margin')
    argument_parser.add_argument('-m',  '--margin',         required=True,  type=float,               help='margin size (micrometer)')
    argument_parser.add_argument('-s',  '--spacing',        required=False, type=float, default=None, help='pixel spacing of processing (micrometer)')
    argument_parser.add_argument('-t',  '--tolerance',      required=False, type=float, default=0.25, help='pixel spacing tolerance (percentage)')
    argument_parser.add_argument('-cd', '--copy_directory', required=False, type=str,   default=None, help='data copy target directory path')
    argument_parser.add_argument('-wd', '--work_directory', required=False, type=str,   default=None, help='intermediate work directory path')
    argument_parser.add_argument('-k',  '--keep_copies',    action='store_true',                      help='keep copied image files')
    argument_parser.add_argument('-w',  '--overwrite',      action='store_true',                      help='overwrite existing results')

    # Parse arguments.
    #
    arguments = vars(argument_parser.parse_args())

    # Collect arguments.
    #
    parsed_input_path = arguments['input']
    parsed_output_path = arguments['output']
    parsed_source_label_value = arguments['label']
    parsed_internal_label_value = arguments['internal']
    parsed_external_label_value = arguments['external']
    parsed_margin_size = arguments['margin']
    parsed_pixel_spacing = arguments['spacing']
    parsed_spacing_tolerance = arguments['tolerance']
    parsed_copy_directory = arguments['copy_directory']
    parsed_work_directory = arguments['work_directory']
    parsed_keep_copies = arguments['keep_copies']
    parsed_overwrite = arguments['overwrite']

    # Print parameters.
    #
    print(argument_parser.description)
    print('Input mask: {path}'.format(path=parsed_input_path))
    print('Output mask: {path}'.format(path=parsed_output_path))
    print('Source label value: {label}'.format(label=parsed_source_label_value))
    print('Target internal label value: {label}'.format(label=parsed_internal_label_value))
    print('Target external label value: {label}'.format(label=parsed_external_label_value))
    print('Margin size: {margin} um'.format(margin=parsed_margin_size))
    print('Processing pixel spacing: {spacing}{measure}'.format(spacing=parsed_pixel_spacing, measure=' um' if parsed_pixel_spacing is not None else ''))
    print('Pixel spacing tolerance: {tolerance}'.format(tolerance=parsed_spacing_tolerance))
    print('Copy directory path: {path}'.format(path=parsed_copy_directory))
    print('Work directory path: {path}'.format(path=parsed_work_directory))
    print('Keep copied files: {flag}'.format(flag=parsed_keep_copies))
    print('Overwrite existing results: {flag}'.format(flag=parsed_overwrite))

    return (parsed_input_path,
            parsed_output_path,
            parsed_source_label_value,
            parsed_internal_label_value,
            parsed_external_label_value,
            parsed_margin_size,
            parsed_pixel_spacing,
            parsed_spacing_tolerance,
            parsed_copy_directory,
            parsed_work_directory,
            parsed_keep_copies,
            parsed_overwrite)

#----------------------------------------------------------------------------------------------------

def main():
    """
    Main function.

    Returns:
        int: Error code.
    """

    # Collect command line arguments.
    #
    input_image_path, output_image_path, label, internal_label, external_label, margin, spacing, tolerance, copy_path, work_path, keep_copied_files, overwrite_flag = collect_arguments()

    # Assemble job triplets: (image, annotation, result mask).
    #
    job_list = assemble_jobs(input_path=input_image_path, output_path=output_image_path, copy_directory=copy_path, work_directory=work_path)

    # Check if there are any identified jobs.
    #
    if job_list:
        # Get a logger object that prints to the console.
        #
        logger = dptloggers.consolelogger(name=__name__, debug=True)

        # Execute jobs.
        #
        successful_items, failed_items = dptregions.add_region_margins_batch(job_list=job_list,
                                                                             source_label=label,
                                                                             target_internal_label=internal_label,
                                                                             target_external_label=external_label,
                                                                             margin_size=margin,
                                                                             pixel_spacing=spacing,
                                                                             spacing_tolerance=tolerance,
                                                                             keep_copies=keep_copied_files,
                                                                             overwrite=overwrite_flag,
                                                                             logger=logger)

        # Print the collection of failed cases.
        #
        if failed_items:
            print('Failed on {count} items:'.format(count=len(failed_items)))
            for path in failed_items:
                print('{path}'.format(path=path))

        error_code = len(failed_items)

    else:
        # Failed to identify any jobs.
        #
        print('No images matched the input filter')

        error_code = -1

    return error_code

#----------------------------------------------------------------------------------------------------


if __name__ == '__main__':

    sys.exit(main())
