"""
This script evaluates the result of a tissue-background segmentation algorithm against a reference.
"""

import digitalpathology.batch.batchsource as dptbatchsource
import digitalpathology.processing.evaluate as dptevaluate

import argparse
import os
import sys

#----------------------------------------------------------------------------------------------------

def assemble_jobs(data_config_path,
                  target_filtering_path,
                  target_segmentation_path,
                  target_comparison_path,
                  target_preview_path,
                  target_table_path,
                  log_path,
                  segmentation_method_name,
                  model_path,
                  purpose_list,
                  category_list,
                  path_overrides,
                  copy_directory,
                  work_directory):
    """
    Collect segmentation, comparision, preview and evaluation job tuples.
    
    Args:
        data_config_path (str): Data configuration file path.
        target_filtering_path (str, None): Target path for reference region filtering. Filtered referece mask files are copied here after processing.
        target_segmentation_path (str, None): Target segmentation file path expression. Segmentation output files copied here after processing.
        target_comparison_path (str, None): Target comparison file path expression. Comparison output files copied here after processing.
        target_preview_path (str, None): Target preview file path expression. Preview output files copied here after processing.
        target_table_path (str): Output summary table path expression.
        log_path (str, None): Log file path expression.
        segmentation_method_name (str): Segmentation method name: 'cnn', 'otsu', 'fesi', or 'threshold'.
        model_path (str): Model file path.
        purpose_list (list): Purpose filter list for the data configuration file.
        category_list (list): Category filter list for the data configuration file.
        path_overrides (dict, None): Path override dictionary for the data configuration file.
        copy_directory (str, None): Copy directory: input files copied here before processing.
        work_directory (str): Work directory path. All intermediate and output files are written here first.

    Returns:
        (list, list, list, list, list, str, str, bool, bool, bool): Filtering jobs, segmentation jobs, comparison jobs, preview jobs, evaluation jobs, result table path, log output path,
            keep filtering files flag, keep segmentation files flag and keep comparison files flag.
    """

    # Assemble job items.
    #
    batch_source = dptbatchsource.BatchSource(source_items=None, mask_spacing=0.0)
    batch_source.load(file_path=data_config_path)
    if path_overrides:
        batch_source.update(path_replacements=path_overrides)

    filtering_job_items = []
    segmentation_job_items = []
    comparison_job_items = []
    preview_job_items = []
    evaluation_job_items = []

    if segmentation_method_name == 'cnn':
        model_basename = os.path.basename(model_path)
        method_name = os.path.splitext(model_basename)[0]
    else:
        method_name = segmentation_method_name

    for source_item in batch_source.items(purpose_id=purpose_list, category_id=category_list, replace=True):
        # Check if image copy is enabled.
        #
        source_image_path = source_item.image_path
        image_basename = os.path.basename(source_image_path)
        image_name = os.path.splitext(image_basename)[0]
        copy_image_path = os.path.join(copy_directory, image_basename.format(image=image_name, method=method_name)) if copy_directory else None

        source_mask_path = source_item.mask_path
        mask_basename = os.path.basename(source_mask_path)
        copy_mask_path = os.path.join(copy_directory, mask_basename.format(image=image_name, method=method_name)) if copy_directory else None

        # Check if reference filtering is saved.
        #
        if target_filtering_path:
            filtering_basename = os.path.basename(target_filtering_path).format(image=image_name, method=method_name)
            filtering_work_path = os.path.join(work_directory, filtering_basename)
            filtering_target_path = os.path.join(os.path.dirname(target_filtering_path), filtering_basename)
        else:
            filtering_basename = '{image}_reference_filtered.tif'.format(image=image_name, method=method_name)
            filtering_work_path = None
            filtering_target_path = os.path.join(work_directory, filtering_basename)

        # Add assembled item to the segmentation job list.
        #
        filtering_job_items.append((source_mask_path, copy_mask_path, filtering_work_path, filtering_target_path))

        # Check if segmentation is saved.
        #
        if target_segmentation_path:
            segmentation_basename = os.path.basename(target_segmentation_path).format(image=image_name, method=method_name)
            segmentation_work_path = os.path.join(work_directory, segmentation_basename)
            segmentation_target_path = os.path.join(os.path.dirname(target_segmentation_path), segmentation_basename)
        else:
            segmentation_basename = '{image}_{method}_segmentation.tif'.format(image=image_name, method=method_name)
            segmentation_work_path = None
            segmentation_target_path = os.path.join(work_directory, segmentation_basename)

        # Add assembled item to the segmentation job list.
        #
        segmentation_job_items.append((source_image_path, None, copy_image_path, None, segmentation_work_path, None, segmentation_target_path, None))

        # Check if comparison is saved.
        #
        if target_comparison_path:
            comparison_basename = os.path.basename(target_comparison_path).format(image=image_name, method=method_name)
            comparison_work_path = os.path.join(work_directory, comparison_basename)
            comparison_target_path = os.path.join(os.path.dirname(target_comparison_path), comparison_basename)
        else:
            comparison_basename = '{image}_{method}_comparison.tif'.format(image=image_name, method=method_name)
            comparison_work_path = None
            comparison_target_path = os.path.join(work_directory, comparison_basename)

        # Add assembled item to the comparison job list.
        #
        comparison_job_items.append((filtering_target_path, segmentation_target_path, filtering_work_path, segmentation_work_path, comparison_work_path, comparison_target_path))

        # Create preview job item if configured.
        #
        if target_preview_path:
            preview_basename = os.path.basename(target_preview_path).format(image=image_name, method=method_name)
            preview_work_path = os.path.join(work_directory, preview_basename)
            preview_target_path = os.path.join(os.path.dirname(target_preview_path), preview_basename)

            preview_job_items.append((source_image_path, comparison_target_path, copy_image_path, comparison_work_path, preview_work_path, preview_target_path))

        # Add assembled item to the evaluation job list.
        #
        evaluation_job_items.append((source_mask_path,
                                     filtering_target_path,
                                     segmentation_target_path,
                                     comparison_target_path,
                                     copy_mask_path,
                                     filtering_work_path,
                                     segmentation_work_path,
                                     comparison_work_path))

    # Apply path replacements on the output table path.
    #
    formatted_table_path = target_table_path.format(method=method_name)

    # Apply path replacements in the output log path.
    #
    formatted_log_path = log_path.format(method=method_name) if log_path else None

    # Check if the result files are to be kept.
    #
    keep_filtered_files = target_filtering_path is not None
    keep_segmentation_files = target_segmentation_path is not None
    keep_comparison_files = target_comparison_path is not None

    # Return the assembled job collections.
    #
    return (filtering_job_items,
            segmentation_job_items,
            comparison_job_items,
            preview_job_items,
            evaluation_job_items,
            formatted_table_path,
            formatted_log_path,
            keep_filtered_files,
            keep_segmentation_files,
            keep_comparison_files)

#----------------------------------------------------------------------------------------------------

def collect_arguments():
    """
    Collect command line arguments.

    Returns:
        (str, float, float, float, str, str, str, str, str, str, str, str, int, int, int, int, float, float, float, float, float, float, list, float, list, list, dict, str, str, bool, bool): The
            parsed command line arguments: input data config path, segmentation pixel spacing, evaluation pixel spacing, preview pixel spacing, output table file path, output reference path, output
            segmentation path, output comparison path, output preview path, log path, method name, model path, grayscale threshold, patch size, reference label, output class label, true object
            detection threshold, false object detection threshold, network confidence threshold, reference region diagonal threshold, segmentation region diagonal threshold, hole diagonal threshold,
            input channels list, pixel spacing tolerance, purpose list for evaluation, category list for evaluation, path overrides in data config file, target directory path for data copy, work
            directory path, channels first flag, and the overwrite flag.
    """

    # Prepare argument value choices.
    #
    method_choices = ['cnn', 'otsu', 'fesi', 'threshold']

    # Configure argument parser.
    #
    argument_parser = argparse.ArgumentParser(description='Calculate image similarity with the selected metric.',
                                              epilog='Use {image}, {method} as replacements for image and model file base-names or method name in the segmentation and preview target '
                                                     'path expression specification. In the output table and log file specifications only the {method} is replaced.')

    argument_parser.add_argument('--input',                        required=True,  type=str,                           help='data config file to process')
    argument_parser.add_argument('--segmentation_spacing',         required=True,  type=float,                         help='input image spacing for segmentation (micrometer)')
    argument_parser.add_argument('--evaluation_spacing',           required=True,  type=float,                         help='mask image spacing for evaluation (micrometer)')
    argument_parser.add_argument('--preview_spacing',              required=False, type=float, default=None,           help='mask image spacing for preview (micrometer)')
    argument_parser.add_argument('--table',                        required=True,  type=str,                           help='output table file path')
    argument_parser.add_argument('--reference_path',               required=False, type=str,   default=None,           help='output filtered reference image path')
    argument_parser.add_argument('--segmentation_path',            required=False, type=str,   default=None,           help='output segmentation image path')
    argument_parser.add_argument('--comparison_path',              required=False, type=str,   default=None,           help='output comparison image path')
    argument_parser.add_argument('--preview_path',                 required=False, type=str,   default=None,           help='output preview image path')
    argument_parser.add_argument('--log',                          required=False, type=str,   default=None,           help='lof file path')
    argument_parser.add_argument('--method',                       required=True,  type=str,   choices=method_choices, help='processing method')
    argument_parser.add_argument('--model',                        required=False, type=str,   default=None,           help='network model for CNN method')
    argument_parser.add_argument('--gray_threshold',               required=False, type=int,   default=None,           help='threshold value for \'threshold\' method')
    argument_parser.add_argument('--patch_size',                   required=False, type=int,   default=1024,           help='processing patch size')
    argument_parser.add_argument('--reference_label',              required=True,  type=int,                           help='reference label')
    argument_parser.add_argument('--template_label',               required=True,  type=int,                           help='template label')
    argument_parser.add_argument('--true_detection_threshold',     required=False, type=float, default=0.8,            help='area ratio threshold for object detection')
    argument_parser.add_argument('--false_detection_threshold',    required=False, type=float, default=0.0,            help='area ratio threshold for object detection')
    argument_parser.add_argument('--confidence',                   required=False, type=float, default=0.5,            help='network confidence for thresholding')
    argument_parser.add_argument('--reference_diagonal_threshold', required=False, type=float, default=0.0,            help='reference region diagonal threshold (micrometer)')
    argument_parser.add_argument('--region_diagonal_threshold',    required=False, type=float, default=0.0,            help='segmentation output region diagonal threshold (micrometer)')
    argument_parser.add_argument('--hole_diagonal_threshold',      required=False, type=float, default=0.0,            help='segmentation output hole diagonal threshold (micrometer)')
    argument_parser.add_argument('--channels',                     required=False, type=str,   default='(0, 1, 2)',    help='list of input channel indices')
    argument_parser.add_argument('--tolerance',                    required=False, type=float, default=0.25,           help='pixel spacing tolerance (percentage)')
    argument_parser.add_argument('--purposes',                     required=False, type=str,   default=None,           help='list of purpose identifiers to use from data configuration file')
    argument_parser.add_argument('--categories',                   required=False, type=str,   default=None,           help='list of category identifiers to use from data configuration file')
    argument_parser.add_argument('--overrides',                    required=False, type=str,   default=None,           help='data config source overrides')
    argument_parser.add_argument('--copy_directory',               required=False, type=str,   default=None,           help='data copy target directory path')
    argument_parser.add_argument('--work_directory',               required=True,  type=str,                           help='work directory path')
    argument_parser.add_argument('--channels_first',               action='store_true',                                help='channels first')
    argument_parser.add_argument('--overwrite',                    action='store_true',                                help='overwrite existing results')

    # Parse arguments.
    #
    arguments = vars(argument_parser.parse_args())

    # Collect arguments.
    #
    parsed_input_path = arguments['input']
    parsed_segmentation_spacing = arguments['segmentation_spacing']
    parsed_evaluation_spacing = arguments['evaluation_spacing']
    parsed_preview_spacing = arguments['preview_spacing']
    parsed_table_path = arguments['table']
    parsed_reference_path = arguments['reference_path']
    parsed_segmentation_path = arguments['segmentation_path']
    parsed_comparison_path = arguments['comparison_path']
    parsed_preview_path = arguments['preview_path']
    parsed_log_path = arguments['log']
    parsed_method = arguments['method']
    parsed_model_path = arguments['model']
    parsed_gray_threshold = arguments['gray_threshold']
    parsed_patch_size = arguments['patch_size']
    parsed_reference_label = arguments['reference_label']
    parsed_template_label = arguments['template_label']
    parsed_true_detection_threshold = arguments['true_detection_threshold']
    parsed_false_detection_threshold = arguments['false_detection_threshold']
    parsed_confidence = arguments['confidence']
    parsed_reference_diagonal_threshold = arguments['reference_diagonal_threshold']
    parsed_region_diagonal_threshold = arguments['region_diagonal_threshold']
    parsed_hole_diagonal_threshold = arguments['hole_diagonal_threshold']
    parsed_channels_str = arguments['channels']
    parsed_spacing_tolerance = arguments['tolerance']
    parsed_purposes_str = arguments['purposes']
    parsed_categories_str = arguments['categories']
    parsed_overrides_str = arguments['overrides']
    parsed_copy_directory = arguments['copy_directory']
    parsed_work_directory = arguments['work_directory']
    parsed_channels_first = arguments['channels_first']
    parsed_overwrite = arguments['overwrite']

    # Evaluate expressions.
    #
    parsed_channels = eval(parsed_channels_str)
    parsed_purposes = eval(parsed_purposes_str) if parsed_purposes_str else None
    parsed_categories = eval(parsed_categories_str) if parsed_categories_str else None
    parsed_overrides = eval(parsed_overrides_str) if parsed_overrides_str else None

    # Print parameters.
    #
    print(argument_parser.description)
    print('Input data configuration path: {path}'.format(path=parsed_input_path))
    print('Segmentation pixel spacing: {spacing} um'.format(spacing=parsed_segmentation_spacing))
    print('Evaluation pixel spacing: {spacing} um'.format(spacing=parsed_evaluation_spacing))
    print('Preview pixel spacing: {spacing} um'.format(spacing=parsed_preview_spacing))
    print('Output table path: {path}'.format(path=parsed_table_path))
    print('Output filtered reference path: {path}'.format(path=parsed_reference_path))
    print('Output segmentation path: {path}'.format(path=parsed_segmentation_path))
    print('Output comparison path: {path}'.format(path=parsed_comparison_path))
    print('Output preview path: {path}'.format(path=parsed_preview_path))
    print('Log path: {path}'.format(path=parsed_log_path))
    print('Method: {method}'.format(method=parsed_method))
    print('Model path: {path}'.format(path=parsed_model_path))
    print('Gray threshold: {threshold}'.format(threshold=parsed_gray_threshold))
    print('Patch size: {size}'.format(size=parsed_patch_size))
    print('Reference label: {label}'.format(label=parsed_reference_label))
    print('Template label: {label}'.format(label=parsed_template_label))
    print('True object detection threshold: {threshold}'.format(threshold=parsed_true_detection_threshold))
    print('False object detection threshold: {threshold}'.format(threshold=parsed_false_detection_threshold))
    print('Network confidence: {confidence}'.format(confidence=parsed_confidence))
    print('Reference region diagonal threshold: {threshold}'.format(threshold=parsed_reference_diagonal_threshold))
    print('Segmentation region diagonal threshold: {threshold}'.format(threshold=parsed_region_diagonal_threshold))
    print('Segmentation hole diagonal threshold: {threshold}'.format(threshold=parsed_hole_diagonal_threshold))
    print('Input channels: {channels}'.format(channels=parsed_channels))
    print('Pixel spacing tolerance: {tolerance}'.format(tolerance=parsed_spacing_tolerance))
    print('Purposes in data config: {purposes}'.format(purposes=parsed_purposes))
    print('Categories in data config: {categories}'.format(categories=parsed_categories))
    print('Path overrides in data config: {overrides}'.format(overrides=parsed_overrides))
    print('Copy directory path: {path}'.format(path=parsed_copy_directory))
    print('Work directory path: {path}'.format(path=parsed_work_directory))
    print('Channels first: {channels_first}'.format(channels_first=parsed_channels_first))
    print('Overwrite existing results: {flag}'.format(flag=parsed_overwrite))

    # Return parsed values.
    #
    return (parsed_input_path,
            parsed_segmentation_spacing,
            parsed_evaluation_spacing,
            parsed_preview_spacing,
            parsed_table_path,
            parsed_reference_path,
            parsed_segmentation_path,
            parsed_comparison_path,
            parsed_preview_path,
            parsed_log_path,
            parsed_method,
            parsed_model_path,
            parsed_gray_threshold,
            parsed_patch_size,
            parsed_reference_label,
            parsed_template_label,
            parsed_true_detection_threshold,
            parsed_false_detection_threshold,
            parsed_confidence,
            parsed_reference_diagonal_threshold,
            parsed_region_diagonal_threshold,
            parsed_hole_diagonal_threshold,
            parsed_channels,
            parsed_spacing_tolerance,
            parsed_purposes,
            parsed_categories,
            parsed_overrides,
            parsed_copy_directory,
            parsed_work_directory,
            parsed_channels_first,
            parsed_overwrite)

#----------------------------------------------------------------------------------------------------

def main():
    """
    Main function.

    Returns:
        int: Error code.
    """

    # Collect command line arguments.
    #
    (input_path,
     segmentation_spacing,
     evaluation_spacing,
     preview_spacing,
     table_path,
     reference_path,
     segmentation_path,
     comparison_path,
     preview_path,
     log_file_path,
     segmentation_method,
     model_file_path,
     gray_threshold,
     input_patch_size,
     reference_label_value,
     template_label_value,
     true_object_detection_threshold,
     false_object_detection_threshold,
     confidence,
     reference_diagonal_threshold,
     region_diagonal_threshold,
     hole_diagonal_threshold,
     channels,
     pixel_spacing_tolerance,
     purposes,
     categories,
     overrides,
     copy_directory_path,
     work_directory_path,
     channels_first_flag,
     overwrite_flag) = collect_arguments()

    # Assemble job items for segmentation, preview calculation and result evaluation.
    #
    (filtering_jobs,
     segmentation_jobs,
     comparison_jobs,
     preview_jobs,
     evaluation_jobs,
     result_table_path,
     result_log_path,
     keep_filtered,
     keep_segmented,
     keep_compared) = assemble_jobs(data_config_path=input_path,
                                    target_filtering_path=reference_path,
                                    target_segmentation_path=segmentation_path,
                                    target_comparison_path=comparison_path,
                                    target_preview_path=preview_path,
                                    target_table_path=table_path,
                                    log_path=log_file_path,
                                    segmentation_method_name=segmentation_method,
                                    model_path=model_file_path,
                                    purpose_list=purposes,
                                    category_list=categories,
                                    path_overrides=overrides,
                                    copy_directory=copy_directory_path,
                                    work_directory=work_directory_path)

    # Evaluate model.
    #
    successful_items, failed_items = dptevaluate.evaluate_model(filtering_jobs=filtering_jobs,
                                                                segmentation_jobs=segmentation_jobs,
                                                                comparison_jobs=comparison_jobs,
                                                                preview_jobs=preview_jobs,
                                                                evaluation_jobs=evaluation_jobs,
                                                                segmentation_pixel_spacing=segmentation_spacing,
                                                                evaluation_pixel_spacing=evaluation_spacing,
                                                                preview_pixel_spacing=preview_spacing,
                                                                output_table_path=result_table_path,
                                                                log_path=result_log_path,
                                                                segmentation_method_name=segmentation_method,
                                                                model_path=model_file_path,
                                                                gray_threshold=gray_threshold,
                                                                patch_size=input_patch_size,
                                                                reference_label=reference_label_value,
                                                                template_label=template_label_value,
                                                                true_detection_threshold=true_object_detection_threshold,
                                                                false_detection_threshold=false_object_detection_threshold,
                                                                confidence_threshold=confidence,
                                                                reference_diagonal_threshold=reference_diagonal_threshold,
                                                                region_diagonal_threshold=region_diagonal_threshold,
                                                                hole_diagonal_threshold=hole_diagonal_threshold,
                                                                input_channels=channels,
                                                                spacing_tolerance=pixel_spacing_tolerance,
                                                                channels_first=channels_first_flag,
                                                                keep_filtering=keep_filtered,
                                                                keep_segmentation=keep_segmented,
                                                                keep_comparison=keep_compared,
                                                                overwrite=overwrite_flag)

    # Print the collection of failed cases.
    #
    if failed_items:
        print('Failed on {count} items:'.format(count=len(failed_items)))
        for path in failed_items:
            print('{path}'.format(path=path))

    # Return error code.
    #
    return len(failed_items)

#----------------------------------------------------------------------------------------------------


if __name__ == '__main__':

    sys.exit(main())
