"""
This module can create and train a network model for classification of slide images.
"""

import digitalpathology.execution.experimentcommander as dptexperimentcommander

import argparse
import sys
import multiprocessing as mp

#----------------------------------------------------------------------------------------------------

def collect_arguments():
    """
    Collect command line arguments.

    Returns:
        (str, str, str, str, str, str, int, list, dict, dict, str, int, bool): The parsed command line arguments: experiment name, parameters file path, data file path, copy directive, progress
            and archive directories, CPU count override, repositories list, the data path and parameter value override maps, file logging level, random seed and the continue experiment flag.
    """

    # Prepare argument value choices.
    #
    logging_level_choices = ['debug', 'info', 'warning', 'error', 'critical']

    # Configure argument parser.
    #
    argument_parser = argparse.ArgumentParser(description='Train network for whole slide image classification.')

    argument_parser.add_argument('-e', '--experiment',          required=True,  type=str,                                                help='experiment name')
    argument_parser.add_argument('-p', '--parameters',          required=True,  type=str,                                                help='input parameters file')
    argument_parser.add_argument('-d', '--data',                required=True,  type=str,                                                help='input data file')
    argument_parser.add_argument('-c', '--copy',                required=False, type=str, default=None,                                  help='data copy source')
    argument_parser.add_argument('-r', '--progress',            required=True,  type=str,                                                help='progress directory')
    argument_parser.add_argument('-a', '--archive',             required=False, type=str, default=None,                                  help='archive directory')
    argument_parser.add_argument('-u', '--cpu',                 required=False, type=int, default=None,                                  help='CPU count override')
    argument_parser.add_argument('-g', '--repositories',        required=False, type=str, default=None,                                  help='additional repositories to report')
    argument_parser.add_argument('-o', '--data_override',       required=False, type=str, default=None,                                  help='data source overrides')
    argument_parser.add_argument('-v', '--parameters_override', required=False, type=str, default=None,                                  help='parameter value overrides')
    argument_parser.add_argument('-l', '--logging_level',       required=False, type=str, default='info', choices=logging_level_choices, help='file logging level')
    argument_parser.add_argument('-s', '--seed',                required=False, type=int, default=None,                                  help='random seed')
    argument_parser.add_argument('-n', '--continue',            action='store_true',                                                     help='continue experiment if possible')

    # Parse arguments.
    #
    arguments = vars(argument_parser.parse_args())

    # Collect arguments.
    #
    parsed_experiment_name = arguments['experiment']
    parsed_param_path = arguments['parameters']
    parsed_data_path = arguments['data']
    parsed_copy_directive_str = arguments['copy']
    parsed_progress_dir_path = arguments['progress']
    parsed_archive_dir_path = arguments['archive']
    parsed_cpu_count = arguments['cpu']
    parsed_repositories_path_list_str = arguments['repositories']
    parsed_data_override_map_str = arguments['data_override']
    parsed_parameters_override_map_str = arguments['parameters_override']
    parsed_logging_level = arguments['logging_level']
    parsed_random_seed = arguments['seed']
    parsed_continue_flag = arguments['continue']

    # Evaluate expressions
    #
    parsed_copy_directive_map = eval(parsed_copy_directive_str) if parsed_copy_directive_str else None
    parsed_repositories_path_list = eval(parsed_repositories_path_list_str) if parsed_repositories_path_list_str else None
    parsed_data_override_map = eval(parsed_data_override_map_str) if parsed_data_override_map_str else None
    parsed_parameters_override_map = eval(parsed_parameters_override_map_str) if parsed_parameters_override_map_str else None

    # Print parameters.
    #
    print(argument_parser.description)
    print('Experiment name: {name}'.format(name=parsed_experiment_name))
    print('Parameters file: {parameters}'.format(parameters=parsed_param_path))
    print('Data file: {data}'.format(data=parsed_data_path))
    print('Copy source: {map}'.format(map=parsed_copy_directive_map))
    print('Progress directory: {progress}'.format(progress=parsed_progress_dir_path))
    print('Archive directory: {move}'.format(move=parsed_archive_dir_path))
    print('CPU count override: {count}'.format(count=parsed_cpu_count))
    print('Repositories to report: {repos}'.format(repos=parsed_repositories_path_list))
    print('Data path overrides: {map}'.format(map=parsed_data_override_map))
    print('Parameter value overrides: {map}'.format(map=parsed_parameters_override_map))
    print('File logging level: {level}'.format(level=parsed_logging_level))
    print('Random seed: {seed}'.format(seed=parsed_random_seed))
    print('Continue experiment: {flag}'.format(flag=parsed_continue_flag))

    # Return parsed values.
    #
    return (parsed_experiment_name,
            parsed_param_path,
            parsed_data_path,
            parsed_copy_directive_map,
            parsed_progress_dir_path,
            parsed_archive_dir_path,
            parsed_cpu_count,
            parsed_repositories_path_list,
            parsed_data_override_map,
            parsed_parameters_override_map,
            parsed_logging_level,
            parsed_random_seed,
            parsed_continue_flag)

#----------------------------------------------------------------------------------------------------

def main():
    """Main function."""

    # Set the default spawning method. Use the 'force=True' parameter in case of error. In theory this setting is not necessary
    #
    mp.set_start_method(method='spawn')

    # Increase recursion limit to be able to save the networks.
    #
    sys.setrecursionlimit(100000)

    # Parse parameters.
    #
    experiment_name, param_path, data_path, copy_directive, progress_dir, archive_dir, cpu_count, repo_list, data_overrides, param_overrides, log_level, seed, continue_experiment = collect_arguments()

    print('')
    print('Executing...')

    # Configure the experiment.
    #
    commander = dptexperimentcommander.ExperimentCommander(experiment_name=experiment_name,
                                                           progress_dir_path=progress_dir,
                                                           archive_dir_path=archive_dir,
                                                           data_file_path=data_path,
                                                           data_overrides=data_overrides,
                                                           data_copy_config=copy_directive,
                                                           param_file_path=param_path,
                                                           param_overrides=param_overrides,
                                                           continue_experiment=continue_experiment,
                                                           random_seed=seed,
                                                           cpu_count_enforce=cpu_count,
                                                           caller_dir_paths=repo_list,
                                                           logging_scope='library',
                                                           file_log_level=log_level,
                                                           console_log_level='info')

    # Execute the learning process.
    #
    commander.execute(network=None)

    print('Exiting')

#----------------------------------------------------------------------------------------------------


if __name__ == '__main__':

    main()
